# -*- coding:utf-8 -*-

from .cfg import DispatchCfg as c
from ..core.callbacks import EarlyStoppingError
from ..core.dispatcher import Dispatcher
from ..core.trial import Trial
from ..utils import logging, fs

logger = logging.get_logger(__name__)


class InProcessDispatcher(Dispatcher):
    def __init__(self, models_dir):
        super(InProcessDispatcher, self).__init__()

        self.models_dir = models_dir
        fs.makedirs(models_dir, exist_ok=True)

    def dispatch(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                 **fit_kwargs):
        retry_limit = c.trial_retry_limit

        trial_no = 1
        retry_counter = 0

        while trial_no <= max_trials:
            try:
                space_sample = hyper_model.searcher.sample()
                if hyper_model.history.is_existed(space_sample):
                    if retry_counter >= retry_limit:
                        logger.info(f'Unable to take valid sample and exceed the retry limit {retry_limit}.')
                        break
                    trial = hyper_model.history.get_trial(space_sample)
                    for callback in hyper_model.callbacks:
                        callback.on_skip_trial(hyper_model, space_sample, trial_no, 'trial_existed',
                                               trial.reward, False, trial.elapsed)
                    retry_counter += 1
                    continue

                if trial_store is not None:
                    trial = trial_store.get(dataset_id, space_sample)
                    if trial is not None:
                        reward = trial.reward
                        elapsed = trial.elapsed
                        trial = Trial(space_sample, trial_no, reward, elapsed)
                        improved = hyper_model.history.append(trial)
                        hyper_model.searcher.update_result(space_sample, reward)
                        for callback in hyper_model.callbacks:
                            callback.on_skip_trial(hyper_model, space_sample, trial_no, 'hit_trial_store', reward,
                                                   improved,
                                                   elapsed)
                        trial_no += 1
                        continue

                for callback in hyper_model.callbacks:
                    callback.on_trial_begin(hyper_model, space_sample, trial_no)

                model_file = '%s/%05d_%s.pkl' % (self.models_dir, trial_no, space_sample.space_id)

                trial = hyper_model._run_trial(space_sample, trial_no, X, y, X_eval, y_eval, cv, num_folds, model_file,
                                               **fit_kwargs)

                if trial.succeeded:
                    improved = hyper_model.history.append(trial)
                    for callback in hyper_model.callbacks:
                        callback.on_trial_end(hyper_model, space_sample, trial_no, trial.reward,
                                              improved, trial.elapsed)
                else:
                    hyper_model.history.append(trial)
                    for callback in hyper_model.callbacks:
                        callback.on_trial_error(hyper_model, space_sample, trial_no)

                if logger.is_info_enabled():
                    msg = f'Trial {trial_no} done, reward: {trial.reward}, ' \
                          f'best_trial_no:{hyper_model.best_trial_no}, best_reward:{hyper_model.best_reward}\n'
                    logger.info(msg)
                if trial_store is not None:
                    trial_store.put(dataset_id, trial)
            except EarlyStoppingError:
                break
            except Exception as e:
                import sys
                import traceback
                msg = f'{">" * 20} Trial {trial_no} failed! {"<" * 20}\n' \
                      + f'{e.__class__.__name__}: {e}\n' \
                      + traceback.format_exc() \
                      + '*' * 50
                logger.error(msg)
            finally:
                trial_no += 1
                retry_counter = 0

        return trial_no



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

import pytest

from hypernets.core.ops import *
from hypernets.core.search_space import *
from hypernets.searchers import GridSearcher
from hypernets.core import EarlyStoppingError


def get_space():
    space = HyperSpace()
    state = np.random.RandomState(9527)

    with space.as_default():
        id1 = Identity(p1=Choice(['a', 'b'], random_state=state),
                       p2=Int(1, 100, random_state=state),
                       p3=Real(0, 1.0, random_state=state))
    return space


class Test_GridSearcher():
    def test_playback_searcher(self):
        searcher = GridSearcher(get_space, n_expansion=2)
        assert searcher.grid == {'Param_Choice_1': ['a', 'b'], 'Param_Int_1': [9, 98],
                                 'Param_Real_1': [0.23, 0.95]}

        assert searcher.all_combinations == [{'Param_Choice_1': 'a', 'Param_Int_1': 9, 'Param_Real_1': 0.23},
                                             {'Param_Choice_1': 'a', 'Param_Int_1': 9,
                                              'Param_Real_1': 0.95},
                                             {'Param_Choice_1': 'a', 'Param_Int_1': 98, 'Param_Real_1': 0.23},
                                             {'Param_Choice_1': 'a', 'Param_Int_1': 98,
                                              'Param_Real_1': 0.95},
                                             {'Param_Choice_1': 'b', 'Param_Int_1': 9, 'Param_Real_1': 0.23},
                                             {'Param_Choice_1': 'b', 'Param_Int_1': 9,
                                              'Param_Real_1': 0.95},
                                             {'Param_Choice_1': 'b', 'Param_Int_1': 98, 'Param_Real_1': 0.23},
                                             {'Param_Choice_1': 'b', 'Param_Int_1': 98,
                                              'Param_Real_1': 0.95}]
        sample1 = searcher.sample()
        assert sample1.vectors == [0, 9, 0.23]
        sample2 = searcher.sample()
        assert sample2.vectors == [0, 9, 0.95]
        sample3 = searcher.sample()
        assert sample3.vectors == [0, 98, 0.23]
        sample4 = searcher.sample()
        assert sample4.vectors == [0, 98, 0.95]
        sample5 = searcher.sample()
        assert sample5.vectors == [1, 9, 0.23]
        sample6 = searcher.sample()
        assert sample6.vectors == [1, 9, 0.95]
        sample7 = searcher.sample()
        assert sample7.vectors == [1, 98, 0.23]
        sample8 = searcher.sample()
        assert sample8.vectors == [1, 98, 0.95]

        with pytest.raises(EarlyStoppingError) as ese:
            searcher.sample()
        assert ese.value.args[0] == 'no more samples.'



import os
import time


def start_broker(host, port):
    from hypernets.dispatchers import run_broker
    from hypernets.dispatchers.process import LocalProcess

    broker_cmd = f'python -m {run_broker.__name__} --host={host} --port={port}'
    broker = LocalProcess(broker_cmd, None, None, None)
    broker.start()

    return broker


def test_grpc_broker_run():
    try:
        from paramiko import SSHClient, AutoAddPolicy
        package_exists = True
    except:
        package_exists = False
    if not package_exists:
        return

    import tempfile
    from hypernets.dispatchers.process import GrpcProcess
    from hypernets.utils.common import generate_id

    broker_host = '127.0.0.1'
    broker_port = 43218
    broker = start_broker(broker_host, broker_port)
    time.sleep(1)

    # run process
    cmd = 'echo 123'
    temp_dir = tempfile.gettempdir()
    test_id = generate_id()
    out_file, err_file = f'{temp_dir}/test_out_{test_id}.out', f'{temp_dir}/test_out_{test_id}.err'
    proc = GrpcProcess(f'{broker_host}:{broker_port}', cmd, None, out_file, err_file)
    proc.run()
    code = proc.exitcode

    with open(out_file, 'r') as f:
        out = f.read()
    with open(err_file, 'r') as f:
        err = f.read()

    # clean down
    os.remove(out_file), os.remove(err_file)
    broker.terminate()  # todo: fix LocalProcess

    # assert
    assert code == 0
    assert out == '123\n'
    assert err.startswith('pid:')



# -*- coding:utf-8 -*-
"""

"""

from .searcher import OptimizeDirection
from .callbacks import Callback, FileStorageLoggingCallback, SummaryCallback, \
    EarlyStoppingCallback, EarlyStoppingError, NotebookCallback, ProgressiveCallback
from .trial import Trial, TrialStore, TrialHistory, DiskTrialStore
from .dispatcher import Dispatcher
from .random_state import set_random_state, get_random_state, randint



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from . import BaseEstimator
from sklearn.metrics import get_scorer

from hypergbm import HyperGBM, CompeteExperiment
from hypergbm.search_space import search_space_general, search_space_feature_gen
from hypernets.core.searcher import OptimizeDirection
from hypernets.experiment import GeneralExperiment, ConsoleCallback
from hypernets.searchers import RandomSearcher, EvolutionSearcher, MCTSSearcher
from hypernets.core import EarlyStoppingCallback, SummaryCallback


class HyperGBMEstimator(BaseEstimator):
    def __init__(self, task, scorer, reward_metric='auc', optimize_direction='max', mode='one-stage', max_trials=30, use_cache=True, earlystop_rounds=30,
                 time_limit=3600, expected_reward=None,
                 drop_feature_with_collinearity=False,
                 search_space_fn=None, ensemble_size=10, use_meta_learner=False, eval_size=0.3,
                 train_test_split_strategy=None,
                 cv=True, num_folds=3, class_balancing=None,
                 retrain_on_wholedata=False,
                 two_stage_importance_selection=True,
                 pseudo_labeling=False,
                 pseudo_labeling_proba_threshold=0.8,
                 pseudo_labeling_resplit=False,
                 **kwargs):
        super(HyperGBMEstimator, self).__init__(task)
        if kwargs.get('name') is not None:
            self.name = kwargs['name']
        else:
            self.name = 'HyperGBM'
        self.scorer = scorer
        self.reward_metric = reward_metric
        self.optimize_direction = optimize_direction
        self.mode = mode
        self.kwargs = kwargs
        self.estimator = None
        self.max_trials = max_trials
        self.use_cache = use_cache
        self.earlystop_rounds = earlystop_rounds
        self.time_limit = time_limit
        self.expected_reward = expected_reward
        self.cv = cv
        self.num_folds = num_folds
        self.search_space_fn = search_space_fn if search_space_fn is not None else lambda: search_space_general(
            early_stopping_rounds=10, verbose=0, class_balancing=class_balancing)
        self.ensemble_size = ensemble_size
        self.experiment = None
        self.use_meta_learner = use_meta_learner
        self.eval_size = eval_size
        self.train_test_split_strategy = train_test_split_strategy
        self.two_stage_importance_selection = two_stage_importance_selection
        self.retrain_on_wholedata = retrain_on_wholedata
        self.pseudo_labeling = pseudo_labeling
        self.pseudo_labeling_proba_threshold = pseudo_labeling_proba_threshold
        self.pseudo_labeling_resplit = pseudo_labeling_resplit
        self.drop_feature_with_collinearity = drop_feature_with_collinearity

    def train(self, X, y, X_test):
        # searcher = MCTSSearcher(self.search_space_fn, use_meta_learner=self.use_meta_learner, max_node_space=10,
        #                         candidates_size=10,
        #                         optimize_direction=OptimizeDirection.Maximize)
        searcher = EvolutionSearcher(self.search_space_fn,
                                     optimize_direction=self.optimize_direction, population_size=30, sample_size=10,
                                     regularized=True, candidates_size=10, use_meta_learner=self.use_meta_learner)
        # searcher = RandomSearcher(lambda: search_space_general(early_stopping_rounds=20, verbose=0),
        #                     optimize_direction=OptimizeDirection.Maximize)
        es = EarlyStoppingCallback(self.earlystop_rounds, self.optimize_direction, time_limit=self.time_limit,
                                   expected_reward=self.expected_reward)

        hk = HyperGBM(searcher, reward_metric=self.reward_metric, cache_dir=f'hypergbm_cache', clear_cache=True,
                      callbacks=[es, SummaryCallback()])

        log_callback = ConsoleCallback()
        self.experiment = CompeteExperiment(hk, X, y, X_test=X_test, eval_size=self.eval_size,
                                            train_test_split_strategy=self.train_test_split_strategy,
                                            cv=self.cv, num_folds=self.num_folds,
                                            callbacks=[],
                                            scorer=get_scorer(self.scorer),
                                            drop_feature_with_collinearity=self.drop_feature_with_collinearity,
                                            drift_detection=True,
                                            n_est_feature_importance=5,
                                            importance_threshold=1e-5,
                                            two_stage_importance_selection=self.two_stage_importance_selection,
                                            ensemble_size=self.ensemble_size,
                                            pseudo_labeling=self.pseudo_labeling,
                                            pseudo_labeling_proba_threshold=self.pseudo_labeling_proba_threshold,
                                            pseudo_labeling_resplit=self.pseudo_labeling_resplit,
                                            retrain_on_wholedata=self.retrain_on_wholedata,
                                            )
        self.estimator = self.experiment.run(use_cache=self.use_cache, max_trials=self.max_trials)

    def predict_proba(self, X):
        return self.estimator.predict_proba(X)

    def predict(self, X):
        return self.estimator.predict(X)



# -*- coding:utf-8 -*-
"""

"""
import math
import os
import pickle
import sys

import numpy as np
import psutil
from dask import dataframe as dd
from joblib import Parallel, delayed
from sklearn import metrics as sk_metrics

from hypernets.utils import const, infer_task_type, logging
from . import dask_ex as dex

logger = logging.get_logger(__name__)

_is_os_windows = sys.platform.find('win') >= 0

_MIN_BATCH_SIZE = 100000

_DASK_METRICS = ('accuracy', 'logloss')


def calc_score(y_true, y_preds, y_proba=None, metrics=('accuracy',), task=const.TASK_BINARY, pos_label=1,
               classes=None, average=None):
    data_list = (y_true, y_proba, y_preds)
    if any(map(dex.is_dask_object, data_list)):
        if all(map(dex.is_dask_object, data_list)) and len(set(metrics).difference(set(_DASK_METRICS))) == 0:
            fn = _calc_score_dask
        else:
            y_true, y_proba, y_preds = \
                [y.compute() if dex.is_dask_object(y) else y for y in data_list]
            fn = _calc_score_sklean
    else:
        fn = _calc_score_sklean

    return fn(y_true, y_preds, y_proba, metrics, task=task, pos_label=pos_label,
              classes=classes, average=average)


def task_to_average(task):
    if task == const.TASK_MULTICLASS:
        average = 'macro'
    else:
        average = 'binary'
    return average


def _calc_score_sklean(y_true, y_preds, y_proba=None, metrics=('accuracy',), task=const.TASK_BINARY, pos_label=1,
                       classes=None, average=None):
    score = {}
    if y_proba is None:
        y_proba = y_preds
    if len(y_proba.shape) == 2 and y_proba.shape[-1] == 1:
        y_proba = y_proba.reshape(-1)
    if len(y_preds.shape) == 2 and y_preds.shape[-1] == 1:
        y_preds = y_preds.reshape(-1)

    if average is None:
        average = task_to_average(task)

    recall_options = dict(average=average, labels=classes)
    if pos_label is not None:
        recall_options['pos_label'] = pos_label

    for metric in metrics:
        if callable(metric):
            score[metric.__name__] = metric(y_true, y_preds)
        else:
            metric_lower = metric.lower()
            if metric_lower == 'auc':
                if len(y_proba.shape) == 2:
                    if task == const.TASK_MULTICLASS:
                        score[metric] = sk_metrics.roc_auc_score(y_true, y_proba, multi_class='ovo', labels=classes)
                    else:
                        score[metric] = sk_metrics.roc_auc_score(y_true, y_proba[:, 1])
                else:
                    score[metric] = sk_metrics.roc_auc_score(y_true, y_proba)
            elif metric_lower == 'accuracy':
                if y_preds is None:
                    score[metric] = 0
                else:
                    score[metric] = sk_metrics.accuracy_score(y_true, y_preds)
            elif metric_lower == 'recall':
                score[metric] = sk_metrics.recall_score(y_true, y_preds, **recall_options)
            elif metric_lower == 'precision':
                score[metric] = sk_metrics.precision_score(y_true, y_preds, **recall_options)
            elif metric_lower == 'f1':
                score[metric] = sk_metrics.f1_score(y_true, y_preds, **recall_options)
            elif metric_lower == 'mse':
                score[metric] = sk_metrics.mean_squared_error(y_true, y_preds)
            elif metric_lower == 'mae':
                score[metric] = sk_metrics.mean_absolute_error(y_true, y_preds)
            elif metric_lower == 'msle':
                score[metric] = sk_metrics.mean_squared_log_error(y_true, y_preds)
            elif metric_lower in {'rmse', 'rootmeansquarederror', 'root_mean_squared_error'}:
                score[metric] = np.sqrt(sk_metrics.mean_squared_error(y_true, y_preds))
            elif metric_lower == 'r2':
                score[metric] = sk_metrics.r2_score(y_true, y_preds)
            elif metric_lower in {'logloss', 'log_loss'}:
                score[metric] = sk_metrics.log_loss(y_true, y_proba, labels=classes)

    return score


def _calc_score_dask(y_true, y_preds, y_proba=None, metrics=('accuracy',), task=const.TASK_BINARY, pos_label=1,
                     classes=None, average=None):
    import dask_ml.metrics as dm_metrics

    def to_array(name, value):
        if value is None:
            return value

        if isinstance(value, (dd.DataFrame, dd.Series)):
            value = value.values

        if len(value.shape) == 2 and value.shape[-1] == 1:
            value = value.reshape(-1)

        value = dex.make_chunk_size_known(value)
        return value

    score = {}

    y_true = to_array('y_true', y_true)
    y_preds = to_array('y_preds', y_preds)
    y_proba = to_array('y_proba', y_proba)

    if y_true.chunks[0] != y_preds.chunks[0]:
        logger.debug(f'rechunk y_preds with {y_true.chunks[0]}')
        y_preds = y_preds.rechunk(chunks=y_true.chunks[0])

    if y_proba is None:
        y_proba = y_preds
    elif y_true.chunks[0] != y_proba.chunks[0]:
        if len(y_proba.chunks) > 1:
            chunks = (y_true.chunks[0],) + y_proba.chunks[1:]
        else:
            chunks = y_true.chunks
        logger.debug(f'rechunk y_proba with {chunks}')
        y_proba = y_proba.rechunk(chunks=chunks)

    for metric in metrics:
        if callable(metric):
            score[metric.__name__] = metric(y_true, y_preds)
        else:
            metric_lower = metric.lower()
            if metric_lower == 'accuracy':
                score[metric] = dm_metrics.accuracy_score(y_true, y_preds)
            elif metric_lower == 'logloss':
                ll = dm_metrics.log_loss(y_true, y_proba, labels=classes)
                if hasattr(ll, 'compute'):
                    ll = ll.compute()
                score[metric] = ll
            else:
                logger.warning(f'unknown metric: {metric}')
    return score


def metric_to_scoring(metric, task=const.TASK_BINARY, pos_label=None):
    assert isinstance(metric, str)

    metric2scoring = {
        'auc': 'roc_auc_ovo',
        'accuracy': 'accuracy',
        'recall': 'recall',
        'precision': 'precision',
        'f1': 'f1',
        'mse': 'neg_mean_squared_error',
        'mae': 'neg_mean_absolute_error',
        'msle': 'neg_mean_squared_log_error',
        'rmse': 'neg_root_mean_squared_error',
        'rootmeansquarederror': 'neg_root_mean_squared_error',
        'root_mean_squared_error': 'neg_root_mean_squared_error',
        'r2': 'r2',
        'logloss': 'neg_log_loss',
    }
    metric2fn = {
        'recall': sk_metrics.recall_score,
        'precision': sk_metrics.precision_score,
        'f1': sk_metrics.f1_score,
    }
    metric_lower = metric.lower()
    if metric_lower not in metric2scoring.keys() and metric_lower not in metric2fn.keys():
        raise ValueError(f'Not found matching scoring for {metric}')

    if metric_lower in metric2fn.keys():
        options = dict(average=task_to_average(task))
        if pos_label is not None:
            options['pos_label'] = pos_label
        scoring = sk_metrics.make_scorer(metric2fn[metric_lower], **options)
    else:
        scoring = sk_metrics.get_scorer(metric2scoring[metric_lower])

    return scoring


def evaluate(estimator, X, y, metrics, *, task=None, pos_label=None, classes=None,
             average=None, threshold=0.5, n_jobs=-1):
    assert classes is None or isinstance(classes, (list, tuple, np.ndarray))
    if isinstance(estimator, str):
        assert os.path.exists(estimator), f'Not found {estimator}'

    if task is None and classes is not None and len(classes) >= 2:
        task = const.TASK_BINARY if len(classes) == 2 else const.TASK_MULTICLASS
    if task is None:
        task, c2 = _detect_task(estimator, y)
        if classes is None or len(classes) < 2:
            classes = c2

    n_jobs = _detect_jobs(X, n_jobs)
    if task in {const.TASK_BINARY, const.TASK_MULTICLASS}:
        proba = predict_proba(estimator, X, n_jobs=n_jobs)
        pred = proba2predict(proba, task=task, threshold=threshold, classes=classes)
    else:
        pred = predict(estimator, X, n_jobs=n_jobs)
        proba = None

    if task == const.TASK_BINARY and pos_label is None:
        pos_label = classes[-1]

    if logger.is_info_enabled():
        logger.info(f'calc_score {metrics}, task={task}, pos_label={pos_label}, classes={classes}, average={average}')
    scores = calc_score(y_true=y, y_preds=pred, y_proba=proba, metrics=metrics,
                        classes=classes, pos_label=pos_label, average=average)

    return scores


def predict_proba(estimator, X, *, n_jobs=-1):
    if isinstance(estimator, str):
        assert os.path.exists(estimator), f'Not found {estimator}'

    n_jobs = _detect_jobs(X, n_jobs)
    if logger.is_info_enabled():
        logger.info(f'predict_proba with n_jobs={n_jobs}')
    proba = _call_predict(estimator, 'predict_proba', X, n_jobs=n_jobs)

    return proba


def predict(estimator, X, *, task=None, classes=None, threshold=0.5, n_jobs=-1):
    assert classes is None or isinstance(classes, (list, tuple, np.ndarray))
    if isinstance(estimator, str):
        assert os.path.exists(estimator), f'Not found {estimator}'

    if task is None and classes is not None and len(classes) >= 2:
        task = const.TASK_BINARY if len(classes) == 2 else const.TASK_MULTICLASS
    if task is None:
        task, c2 = _detect_task(estimator, None)
        if classes is None or len(classes) < 2:
            classes = c2

    if task == const.TASK_REGRESSION:
        if logger.is_info_enabled():
            logger.info(f'predict with n_jobs={n_jobs}')
        pred = _call_predict(estimator, 'predict', X, n_jobs=n_jobs)
    else:
        proba = predict_proba(estimator, X, n_jobs=n_jobs)
        if task is None and (len(proba.shape) < 2 or proba.shape[1] == 1):
            task = const.TASK_REGRESSION
        pred = proba2predict(proba, task=task, threshold=threshold, classes=classes)
    return pred


def proba2predict(proba, *, task=None, threshold=0.5, classes=None):
    assert len(proba.shape) <= 2

    def is_one_dim(x):
        return len(x.shape) == 1 or (len(x.shape) == 2 and x.shape[1] == 1)

    if logger.is_info_enabled():
        logger.info(f'proba2predict with task={task}, classes={classes}, threshold={threshold}')

    if len(proba.shape) == 0:  # empty
        return proba

    if task == const.TASK_BINARY and is_one_dim(proba):
        proba = dex.fix_binary_predict_proba_result(proba)

    if task == const.TASK_REGRESSION or is_one_dim(proba):  # regression
        return proba

    if proba.shape[-1] > 2:  # multiclass
        predict = proba.argmax(axis=-1)
    else:  # binary
        predict = (proba[:, -1] > threshold).astype(np.int32)

    if classes is not None:
        if dex.is_dask_object(predict):
            predict = dex.da.take(np.array(classes), predict, axis=0)
        else:
            predict = np.take(np.array(classes), predict, axis=0)

    return predict


def _detect_jobs(X, n_jobs):
    if callable(getattr(X, 'compute', None)):  # dask data frame
        return 1

    assert X.shape[0] > 0, f'Not found data.'

    if n_jobs <= 0:
        n_jobs = math.ceil(X.shape[0] / _MIN_BATCH_SIZE)

    cores = psutil.cpu_count()
    if n_jobs > cores:
        n_jobs = cores

    return n_jobs


def _detect_task(estimator, y):
    if isinstance(estimator, str):
        with open(estimator, 'rb') as f:
            estimator = pickle.load(f)

    task = None
    classes = None

    if task is None and hasattr(estimator, 'task'):
        task = getattr(estimator, 'task', None)

    if task is None and type(estimator).__name__.find('Pipeline') >= 0 and hasattr(estimator, 'steps'):
        task = getattr(estimator.steps[-1][1], 'task', None)

    if hasattr(estimator, 'classes_'):
        classes = getattr(estimator, 'classes_', None)
        if not (isinstance(classes, (list, tuple, np.ndarray)) and len(classes) > 1):
            classes = None
        if task is None and classes is not None:
            task = const.TASK_BINARY if len(classes) == 2 else const.TASK_MULTICLASS

    if classes is None and type(estimator).__name__.find('Pipeline') >= 0 and hasattr(estimator, 'steps'):
        classes = getattr(estimator.steps[-1][1], 'classes_', None)
        if not (isinstance(classes, (list, tuple, np.ndarray)) and len(classes) > 1):
            classes = None
        if task is None and classes is not None:
            task = const.TASK_BINARY if len(classes) == 2 else const.TASK_MULTICLASS

    if task is None and y is not None:
        task, c2 = infer_task_type(y)
        if classes is None:
            classes = c2

    return task, classes


def _load_and_run(estimator, fn_name, df):
    if isinstance(estimator, str):
        logger.info(f'load estimator {estimator}')
        with open(estimator, 'rb') as f:
            estimator = pickle.load(f)

    fn = getattr(estimator, fn_name)
    assert callable(fn)

    logger.info(f'call {fn_name}')
    result = fn(df)

    return result


def _call_predict(estimator, fn_name, df, n_jobs=1):
    if n_jobs > 1:
        batch_size = math.ceil(df.shape[0] / n_jobs)
        df_parts = [df[i:i + batch_size].copy() for i in range(df.index.start, df.index.stop, batch_size)]
        options = dict(backend='multiprocessing') if _is_os_windows else dict(prefer='processes')
        pss = Parallel(n_jobs=n_jobs, **options)(delayed(_load_and_run)(estimator, fn_name, x) for x in df_parts)

        if len(pss[0].shape) > 1:
            result = np.vstack(pss)
        else:
            result = np.hstack(pss)
    else:
        result = _load_and_run(estimator, fn_name, df)

    return result



# -*- coding:utf-8 -*-
"""

"""
import math
from datetime import datetime

import dask.dataframe as dd
import pandas as pd
import pytest
from sklearn.pipeline import Pipeline

from hypernets.tabular import dask_ex as dex
from hypernets.tabular.column_selector import column_object_category_bool, column_number_exclude_timedelta
from hypernets.tabular.dataframe_mapper import DataFrameMapper
from hypernets.tabular.datasets import dsutils
from hypernets.tabular.feature_generators import FeatureGenerationTransformer
from hypernets.tests.tabular.dask_transofromer_test import setup_dask
from hypernets.utils import logging

logger = logging.getLogger(__name__)


def general_preprocessor():
    cat_transformer = Pipeline(
        steps=[('imputer_cat', dex.SimpleImputer(strategy='constant', fill_value='')),
               ('encoder', dex.OrdinalEncoder())])
    num_transformer = Pipeline(
        steps=[('imputer_num', dex.SimpleImputer(strategy='mean')),
               ('scaler', dex.StandardScaler())])

    preprocessor = DataFrameMapper(features=[(column_object_category_bool, cat_transformer),
                                             (column_number_exclude_timedelta, num_transformer)],
                                   input_df=True,
                                   df_out=True)
    return preprocessor


class Test_FeatureGeneratorWithDask:
    @classmethod
    def setup_class(cls):
        setup_dask(cls)

    def test_pipeline(self):
        df = dsutils.load_bank()
        df.drop(['id'], axis=1, inplace=True)
        ddf = dex.dd.from_pandas(df.head(100), npartitions=2)
        X_train, X_test = dex.train_test_split(ddf, test_size=0.2, random_state=42)
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=['cross_categorical'],
                                           categories_cols=column_object_category_bool(X_train))
        preprocessor = general_preprocessor()
        pipe = Pipeline(steps=[('feature_gen', ftt), ('processor', preprocessor)])
        X_t = pipe.fit_transform(X_train)
        X_t = X_t.compute()
        assert X_t.shape[1] == 62

    def test_in_dataframe_mapper(self):
        df = dsutils.load_bank()
        df.drop(['id'], axis=1, inplace=True)
        ddf = dex.dd.from_pandas(df.head(100), npartitions=2)
        X_train, X_test = dex.train_test_split(ddf, test_size=0.2, random_state=42)
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=['cross_categorical'],
                                           categories_cols=column_object_category_bool(X_train))
        dfm = DataFrameMapper(features=[(X_train.columns.to_list(), ftt)],
                              input_df=True,
                              df_out=True)
        X_t = dfm.fit_transform(X_train)
        X_t = X_t.compute()
        assert X_t.shape[1] == 62

    def test_feature_tools_categorical_cross(self):
        df = dsutils.load_bank()
        df.drop(['id'], axis=1, inplace=True)
        ddf = dex.dd.from_pandas(df.head(100), npartitions=2)
        X_train, X_test = dex.train_test_split(ddf, test_size=0.2, random_state=42)
        cat_cols = column_object_category_bool(X_train)
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=['cross_categorical'],
                                           categories_cols=cat_cols)
        ftt.fit(X_train)
        x_t = ftt.transform(X_train)
        columns = set(x_t.columns.to_list())
        for i_left in range(len(cat_cols) - 1):
            for i_right in range(i_left + 1, len(cat_cols)):
                assert f'CROSS_CATEGORICAL_{cat_cols[i_left]}__{cat_cols[i_right]}' in columns \
                       or f'CROSS_CATEGORICAL_{cat_cols[i_right]}__{cat_cols[i_left]}' in columns

    def test_feature_tools_transformer(self):
        df = dsutils.load_bank()
        df.drop(['id'], axis=1, inplace=True)
        y = df.pop('y')
        ddf = dex.dd.from_pandas(df.head(100), npartitions=2)
        X_train, X_test = dex.train_test_split(ddf, test_size=0.2, random_state=42)
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=['add_numeric', 'divide_numeric'])
        ftt.fit(X_train)
        x_t = ftt.transform(X_train)
        assert x_t is not None

    # def test_feature_selection(self):
    #     df = dsutils.load_bank().head(1000)
    #     df.drop(['id'], axis=1, inplace=True)
    #     y = df.pop('y')
    #     ftt = FeatureGenerationTransformer(task='binary',
    #                                        trans_primitives=['add_numeric', 'divide_numeric', 'cross_categorical'],
    #                                        categories_cols=column_object_category_bool(df))
    #     ftt.fit(df)
    #     x_t = ftt.transform(df)
    #
    #     fst = FeatureSelectionTransformer('binary', ratio_select_cols=0.2, reserved_cols=ftt.original_cols)
    #     fst.fit(x_t, y)
    #     assert len(fst.scores_.items()) == 99
    #     assert len(fst.columns_) == 35
    #     x_t2 = fst.transform(x_t)
    #     assert x_t2.shape[1] == 35

    def test_category_datetime_text(self):
        df = dsutils.load_movielens()
        df['genres'] = df['genres'].apply(lambda s: s.replace('|', ' '))
        df['timestamp'] = df['timestamp'].apply(datetime.fromtimestamp)
        ddf = dd.from_pandas(df, npartitions=2)
        ftt = FeatureGenerationTransformer(task='binary', text_cols=['title'], categories_cols=['gender', 'genres'])
        x_t = ftt.fit_transform(ddf)
        xt_columns = x_t.columns.to_list()
        assert 'CROSS_CATEGORICAL_gender__genres' in xt_columns
        assert 'TFIDF__title____0__' in xt_columns
        assert 'DAY__timestamp__' in xt_columns

    def test_latlong(self):
        df = pd.DataFrame()
        df['latitude'] = [51.52, 9.93, 37.38]
        df['longitude'] = [-0.17, 76.25, -122.08]
        df['latlong'] = df[['latitude', 'longitude']].apply(tuple, axis=1)
        df['latitude2'] = [51.22, 9.22, 37.22]
        df['longitude2'] = [-0.22, 76.22, -122.22]
        df['latlong2'] = df[['latitude2', 'longitude2']].apply(tuple, axis=1)
        df = dd.from_pandas(df, npartitions=1)
        ftt = FeatureGenerationTransformer(latlong_cols=['latlong', 'latlong2'])
        x_t = ftt.fit_transform(df)
        print(x_t.head(3))
        assert 'GEOHASH__latlong__' in x_t.columns.to_list()

    # def test_feature_generation_with_selection(self):
    #     df = dsutils.load_bank().head(1000)
    #     df.drop(['id'], axis=1, inplace=True)
    #     y = df.pop('y')
    #     ftt = FeatureGenerationTransformer(task='binary',
    #                                        trans_primitives=['add_numeric', 'divide_numeric', 'cross_categorical'],
    #                                        categories_cols=column_object_category_bool(df),
    #                                        feature_selection_args={'ratio_select_cols': 0.2})
    #     with pytest.raises(AssertionError) as err:
    #         ftt.fit(df)
    #         assert err.value == '`y` must be provided for feature selection.'
    #     ftt.fit(df, y)
    #     x_t = ftt.transform(df)
    #     assert x_t.shape[1] == 35

    @pytest.mark.parametrize('fix_input', [True, False])
    def test_fix_input(self, fix_input: bool):
        df = pd.DataFrame(data={"x1": [None, 2, 3], 'x2': [4, 5, 6]})
        df = dex.dd.from_pandas(df, npartitions=1)
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=['add_numeric', 'divide_numeric'],
                                           fix_input=fix_input)
        ftt.fit(df)
        x_t = ftt.transform(df)
        x_t = x_t.compute()
        assert "x1__A__x2" in x_t
        assert "x1__D__x2" in x_t

        if fix_input is True:
            # should no NaN value not only input nor output
            assert not math.isnan(x_t["x1"][0])
            assert not math.isnan(x_t["x1__D__x2"][0])
        else:
            # x1 is NaN, it's children is NaN too.
            assert math.isnan(x_t["x1"][0])
            assert math.isnan(x_t["x1__D__x2"][0])

    def test_datetime_derivation(self):

        df = pd.DataFrame(data={"x1": [datetime.now()]})
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=["year", "month", "week"])
        ftt.fit(df)

        x_t = ftt.transform(df)
        assert "YEAR__x1__" in x_t
        assert "MONTH__x1__" in x_t
        assert "WEEK__x1__" in x_t

    def test_persist(self, tmp_path: str):
        from os import path as P
        tmp_path = P.join(tmp_path, 'fft.pkl')

        df = pd.DataFrame(data={"x1": [datetime.now()]})
        ftt = FeatureGenerationTransformer(task='binary', trans_primitives=["year", "month", "week"])
        ftt.fit(df)
        import pickle

        with open(tmp_path, 'wb') as f:
            pickle.dump(ftt, f)

        with open(tmp_path, 'rb') as f:
            ftt1 = pickle.load(f)

        x_t = ftt1.transform(df)
        assert "YEAR__x1__" in x_t
        assert "MONTH__x1__" in x_t
        assert "WEEK__x1__" in x_t



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

import copy
import time
import pandas as pd
from hypernets.conf import configure, Configurable, String, Int as cfg_int
from hypernets.core import TrialHistory, Trial, EarlyStoppingError
from hypernets.core.ops import Identity, HyperInput
from hypernets.core.search_space import HyperSpace, ParameterSpace
from hypernets.searchers import EvolutionSearcher, RandomSearcher, MCTSSearcher, GridSearcher, get_searcher_cls, \
    Searcher
from hypernets.utils import logging
from hypernets.utils.common import isnotebook
from IPython.display import clear_output, display, update_display

logger = logging.get_logger(__name__)


@configure()
class ParamSearchCfg(Configurable):
    work_dir = String(help='storage directory path to store running data.').tag(config=True)
    trial_retry_limit = cfg_int(1000, min=1, help='maximum retry number to run trial.').tag(config=True)


def func_space(func):
    space = HyperSpace()
    with space.as_default():
        params = {name: copy.copy(v) for name, v in zip(func.__code__.co_varnames, func.__defaults__) if
                  isinstance(v, ParameterSpace)}
        for _, v in params.items():
            v.attach_to_space(space, v.name)
        input = HyperInput()
        id1 = Identity(**params)(input)
        space.set_outputs([id1])
    return space


def build_searcher(cls, func, optimize_direction='min'):
    cls = get_searcher_cls(cls)
    search_space_fn = lambda: func_space(func)

    if cls == EvolutionSearcher:
        s = cls(search_space_fn, optimize_direction=optimize_direction,
                population_size=30, sample_size=10, candidates_size=10,
                regularized=True, use_meta_learner=True)
    elif cls == MCTSSearcher:
        s = MCTSSearcher(search_space_fn, optimize_direction=optimize_direction, max_node_space=10)
    elif cls == RandomSearcher:
        s = cls(search_space_fn, optimize_direction=optimize_direction)
    elif cls == GridSearcher:
        s = cls(search_space_fn, optimize_direction=optimize_direction)
    else:
        s = cls(search_space_fn, optimize_direction=optimize_direction)
    return s


def search_params(func, searcher='Grid', max_trials=100, optimize_direction='min', clear_logs=False, **func_kwargs):
    if not isinstance(searcher, Searcher):
        searcher = build_searcher(searcher, func, optimize_direction)
    retry_limit = ParamSearchCfg.trial_retry_limit
    trial_no = 1
    retry_counter = 0
    history = TrialHistory(optimize_direction)

    current_trial_display_id = None
    trials_display_id = None
    while trial_no <= max_trials:
        try:
            space_sample = searcher.sample()

            if isnotebook():
                if clear_logs:
                    clear_output()
                    current_trial_display_id = None
                    trials_display_id = None
                    
                if current_trial_display_id is None:
                    display({'text/markdown': '#### Current Trial:'}, raw=True, include=['text/markdown'])
                    handle = display(space_sample, display_id=True)
                    if handle is not None:
                        current_trial_display_id = handle.display_id
                else:
                    update_display(space_sample, display_id=current_trial_display_id)
                df_best_trials = pd.DataFrame([
                    (t.trial_no, t.reward, t.elapsed, t.space_sample.vectors) for t in history.get_top(100)],
                    columns=['Trial No.', 'Reward', 'Elapsed', 'Space Vector'])
                if trials_display_id is None:
                    display({'text/markdown': '#### Top trials:'}, raw=True, include=['text/markdown'])
                    handle = display(df_best_trials, display_id=True)
                    if handle is not None:
                        trials_display_id = handle.display_id
                else:
                    update_display(df_best_trials, display_id=trials_display_id)

            if history.is_existed(space_sample):
                if retry_counter >= retry_limit:
                    logger.info(f'Unable to take valid sample and exceed the retry limit {retry_limit}.')
                    break
                retry_counter += 1
                continue

            ps = space_sample.get_assigned_params()
            params = {p.alias.split('.')[-1]: p.value for p in ps}
            func_params = func_kwargs.copy()
            func_params.update(params)

            trial_start = time.time()
            last_reward = func(**func_params)
            elapsed = time.time() - trial_start

            trial = Trial(space_sample, trial_no, last_reward, elapsed)
            if last_reward != 0:  # success
                improved = history.append(trial)
            if logger.is_info_enabled():
                best = history.get_best()
                msg = f'Trial {trial_no} done, reward: {trial.reward}, ' \
                      f'best_trial_no:{best.trial_no}, best_reward:{best.reward}\n'
                logger.info(msg)
        except EarlyStoppingError:
            break
            # TODO: early stopping
        except Exception as e:
            import sys
            import traceback
            msg = f'{">" * 20} Trial {trial_no} failed! {"<" * 20}\n' \
                  + f'{e.__class__.__name__}: {e}\n' \
                  + traceback.format_exc() \
                  + '*' * 50
            logger.error(msg)
        finally:
            trial_no += 1
            retry_counter = 0
    return history



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from .mcts_core import *
from ..core.searcher import Searcher, OptimizeDirection


class MCTSSearcher(Searcher):
    """
    References
    ----------
        Wang, Linnan, et al. "Alphax: exploring neural architectures with deep neural networks and monte carlo tree search." arXiv preprint arXiv:1903.11059 (2019).
        Browne, Cameron B., et al. "A survey of monte carlo tree search methods." IEEE Transactions on Computational Intelligence and AI in games 4.1 (2012): 1-43.
    """

    def __init__(self, space_fn, policy=None, max_node_space=10, candidates_size=10,
                 optimize_direction=OptimizeDirection.Minimize, use_meta_learner=True, space_sample_validation_fn=None):
        """
        :param space_fn: Callable
            A search space function which when called returns a `HyperSpace` object.
        :param policy: hypernets.searchers.mcts_core.BasePolicy, (default=None)
            The policy for *Selection* and *Backpropagation* phases, `UCT` by default.
        :param max_node_space: int, (default=10)
            Maximum space for node expansion
        :param candidates_size: int, (default=10)
            The number of samples for the meta-learner to evaluate candidate paths when roll out
        :param optimize_direction: 'min' or 'max', (default='min')
            Whether the search process is approaching the maximum or minimum reward value
        :param use_meta_learner: bool, (default=True)
            Meta-learner aims to evaluate the performance of unseen samples based on previously evaluated samples. It provides a practical solution to accurately estimate a search branch with many simulations without involving the actual training
        :param space_sample_validation_fn: Callable or None, (default=None)
            Used to verify the validity of samples from the search space, and can be used to add specific constraint rules to the search space to reduce the size of the space
        """
        if policy is None:
            policy = UCT()
        self.tree = MCTree(space_fn, policy, max_node_space=max_node_space)
        Searcher.__init__(self, space_fn, optimize_direction, use_meta_learner=use_meta_learner,
                          space_sample_validation_fn=space_sample_validation_fn)
        self.nodes_map = {}
        self.candidates_size = candidates_size

    @property
    def max_node_space(self):
        return self.tree.max_node_space

    def parallelizable(self):
        return self.use_meta_learner and self.meta_learner is not None

    def sample(self):
        # print('Sample')
        _, best_node = self.tree.selection_and_expansion()
        # print(f'Sample: {best_node.info()}')

        if self.use_meta_learner and self.meta_learner is not None:
            space_sample, candidate_sim_score, candidates_avg_score = self._select_best_candidate(best_node)
            # support for parallelize sampling
            self.tree.back_propagation(best_node, candidates_avg_score, is_simulation=True)
        else:
            space_sample = self._roll_out(best_node)

        self.nodes_map[space_sample.space_id] = best_node
        return space_sample

    def _roll_out(self, node):
        def sample():
            space_sample = self.tree.node_to_space(node)
            space_sample = self.tree.roll_out(space_sample, node)
            return space_sample

        space_sample = self._sample_and_check(sample_fn=sample)
        return space_sample

    def _select_best_candidate(self, node):
        candidates = []
        scores = []
        for i in range(self.candidates_size):
            candidate = self._roll_out(node)
            candidates.append(candidate)
            scores.append(self.meta_learner.predict(candidate, 0.5))
        index = np.argmax(scores)
        candidate_sim_score = scores[index]
        candidates_avg_score = np.average(scores)
        # print(f'selected candidates scores:{scores}, argmax:{index}')
        return candidates[index], candidate_sim_score, candidates_avg_score

    def get_best(self):
        raise NotImplementedError

    def update_result(self, space_sample, result):
        best_node = self.nodes_map[space_sample.space_id]
        # print(f'Update result: space:{space_sample.space_id}, result:{result}, node:{best_node.info()}')
        self.tree.back_propagation(best_node, result)
        # print(f'After back propagation: {best_node.info()}')
        # print('\n\n')
        if self.use_meta_learner and self.meta_learner is not None:
            assert self.meta_learner is not None
            self.meta_learner.new_sample(space_sample)

    def summary(self):
        return str(self.tree.root)

    def reset(self):
        raise NotImplementedError

    def export(self):
        raise NotImplementedError



# -*- coding:utf-8 -*-
"""

"""
import copy

import numpy as np
from sklearn.model_selection import KFold, StratifiedKFold


class Estimator():
    def __init__(self, space_sample, task='binary', discriminator=None):
        self.space_sample = space_sample
        self.task = task
        self.discriminator = discriminator

    def set_discriminator(self, discriminator):
        self.discriminator = discriminator

    def summary(self):
        raise NotImplementedError

    def fit(self, X, y, **kwargs):
        raise NotImplementedError

    def fit_cross_validation(self, X, y, stratified=True, num_folds=3,
                             shuffle=False, random_state=9527, metrics=None):
        raise NotImplementedError

    def predict(self, X, **kwargs):
        raise NotImplementedError

    def predict_proba(self, X, **kwargs):
        raise NotImplementedError

    def evaluate(self, X, y, metrics=None, **kwargs):
        raise NotImplementedError

    def save(self, model_file):
        raise NotImplementedError

    @staticmethod
    def load(model_file):
        raise NotImplementedError

    def proba2predict(self, proba, proba_threshold=0.5):
        if self.task == 'regression':
            return proba
        if proba.shape[-1] > 2:
            predict = proba.argmax(axis=-1)
        elif proba.shape[-1] == 2:
            predict = (proba[:, 1] > proba_threshold).astype('int32')
        else:
            predict = (proba > proba_threshold).astype('int32')
        return predict

    def get_iteration_scores(self):
        return []


class CrossValidationEstimator():
    def __init__(self, base_estimator, task, num_folds=3, stratified=False, shuffle=False, random_state=None):
        self.base_estimator = base_estimator
        self.num_folds = num_folds
        self.stratified = stratified
        self.shuffle = shuffle
        self.random_state = random_state
        self.task = task
        self.oof_ = None
        self.classes_ = None
        self.estimators_ = []

    def fit(self, X, y, **kwargs):
        self.oof_ = None
        self.estimators_ = []
        if self.stratified and self.task == 'binary':
            iterators = StratifiedKFold(n_splits=self.num_folds, shuffle=self.shuffle, random_state=self.random_state)
        else:
            iterators = KFold(n_splits=self.num_folds, shuffle=self.shuffle, random_state=self.random_state)

        y = np.array(y)
        sample_weight = kwargs.get('sample_weight')

        for n_fold, (train_idx, valid_idx) in enumerate(iterators.split(X, y)):
            x_train_fold, y_train_fold = X.iloc[train_idx], y[train_idx]
            x_val_fold, y_val_fold = X.iloc[valid_idx], y[valid_idx]

            kwargs['eval_set'] = [(x_val_fold, y_val_fold)]
            if sample_weight is not None:
                sw_fold = sample_weight[train_idx]
                kwargs['sample_weight'] = sw_fold
            fold_est = copy.deepcopy(self.base_estimator)
            fold_est.fit(x_train_fold, y_train_fold, **kwargs)
            if self.classes_ is None:
                self.classes_ = fold_est.classes_
            if self.task == 'regression':
                proba = fold_est.predict(x_val_fold)
            else:
                proba = fold_est.predict_proba(x_val_fold)

            if self.oof_ is None:
                if len(proba.shape) == 1:
                    self.oof_ = np.zeros(y.shape, proba.dtype)
                else:
                    self.oof_ = np.zeros((y.shape[0], proba.shape[-1]), proba.dtype)
            self.oof_[valid_idx] = proba
            self.estimators_.append(fold_est)

        return self

    def predict_proba(self, X):
        proba_sum = None
        for est in self.estimators_:
            proba = est.predict_proba(X)
            if proba_sum is None:
                proba_sum = proba
            else:
                proba_sum += proba
        return proba_sum / len(self.estimators_)

    def predict(self, X):
        if self.task == 'regression':
            pred_sum = None
            for est in self.estimators_:
                pred = est.predict(X)
                if pred_sum is None:
                    pred_sum = pred
                else:
                    pred_sum += pred
            return pred_sum / len(self.estimators_)
        elif self.task == 'binary':
            proba = self.predict_proba(X)
            pred = self.proba2predict(proba)
            pred = np.array(self.classes_).take(pred, axis=0)
            return pred

    def proba2predict(self, proba, proba_threshold=0.5):
        assert len(proba.shape) <= 2
        if self.task == 'regression':
            return proba
        if len(proba.shape) == 2:
            if proba.shape[-1] > 2:
                predict = proba.argmax(axis=-1)
            else:
                predict = (proba[:, -1] > proba_threshold).astype('int32')
        else:
            predict = (proba > proba_threshold).astype('int32')

        return predict



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

TASK_AUTO = 'auto'
TASK_BINARY = 'binary'
TASK_MULTICLASS = 'multiclass'
TASK_REGRESSION = 'regression'
TASK_MULTILABEL = 'multilabel'


COLUMNNAME_POSTFIX_DISCRETE = '_discrete'
COLUMNNAME_POSTFIX_CATEGORIZE = '_cat'

# DATATYPE_TENSOR_FLOAT = 'float32'
# DATATYPE_PREDICT_CLASS = 'int32'
DATATYPE_LABEL = 'int16'


# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from hypernets.utils.logging import get_logger

from ._base import BaseDiscriminator, get_percentile_score

logger = get_logger(__name__)


class PercentileDiscriminator(BaseDiscriminator):
    def __init__(self, percentile, min_trials=5, min_steps=5, stride=1, history=None, optimize_direction='min'):
        assert 0.0 <= percentile <= 100.0, f'Percentile which must be between 0 and 100 inclusive. got {percentile}'

        BaseDiscriminator.__init__(self, min_trials, min_steps, stride, history, optimize_direction)
        self.percentile = percentile

    def _is_promising(self, iteration_trajectory, group_id):
        n_step = len(iteration_trajectory) - 1
        percentile_score = get_percentile_score(self.history, n_step, group_id, self.percentile, self._sign)
        current_trial_score = iteration_trajectory[-1]
        result = current_trial_score * self._sign > percentile_score * self._sign
        if not result and logger.is_info_enabled():
            logger.info(f'direction:{self.optimize_direction}, promising:{result}, '
                        f'percentile_score:{percentile_score}, current_trial_score:{current_trial_score}, '
                        f'trajectory size:{n_step + 1}')
        return result


class ProgressivePercentileDiscriminator(BaseDiscriminator):
    def __init__(self, percentile_list, min_trials=5, min_steps=5, stride=1, history=None, optimize_direction='min'):
        assert len(percentile_list) > 0, 'percentile list has at least one element'
        assert all([0.0 <= percentile <= 100.0 for percentile in percentile_list]), \
            f'Percentile which must be between 0 and 100 inclusive. got {percentile_list}'

        BaseDiscriminator.__init__(self, min_trials, min_steps, stride, history, optimize_direction)
        self.percentile_list = percentile_list

    def _is_promising(self, iteration_trajectory, group_id):
        n_step = len(iteration_trajectory) - 1
        progress = (n_step - self.min_steps) // self.stride

        if progress < len(self.percentile_list):
            percentile = self.percentile_list[progress]
        else:
            percentile = self.percentile_list[-1]

        percentile_score = get_percentile_score(self.history, n_step, group_id, percentile, self._sign)
        current_trial_score = iteration_trajectory[-1]
        result = current_trial_score * self._sign > percentile_score * self._sign
        if not result and logger.is_info_enabled():
            logger.info(f'progress:{progress}, percentile:{percentile}, is promising:{result}, '
                        f'percentile_score:{percentile_score}, current_trial_score:{current_trial_score}, '
                        f'trajectory size:{n_step + 1}')
        return result



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from ._doc_lens import DocLens
from ._fsutils import filesystem as fs
from ._tic_tok import tic_toc, report as tic_toc_report, report_as_dataframe as tic_toc_report_as_dataframe
from .common import generate_id, combinations, isnotebook, Counter, to_repr, get_params
from .common import infer_task_type, hash_data, hash_dataframe, load_data, load_module
from ._estimators import load_estimator, save_estimator



# -*- coding:utf-8 -*-
"""
Adapted from: https://github.com/scikit-learn-contrib/sklearn-pandas
1. Fix the problem of confusion of column names
2. Support `columns` is a callable object
"""
import contextlib
import sys

import numpy as np
import pandas as pd
import six
from dask import array as da
from dask import dataframe as dd
from scipy import sparse as _sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import _name_estimators, Pipeline
from sklearn.utils import tosequence

from hypernets.utils import logging
from . import dask_ex as dex

logger = logging.get_logger(__name__)


def _call_fit(fit_method, X, y=None, **kwargs):
    """
    helper function, calls the fit or fit_transform method with the correct
    number of parameters

    fit_method: fit or fit_transform method of the transformer
    X: the data to fit
    y: the target vector relative to X, optional
    kwargs: any keyword arguments to the fit method

    return: the result of the fit or fit_transform method

    WARNING: if this function raises a TypeError exception, test the fit
    or fit_transform method passed to it in isolation as _call_fit will not
    distinguish TypeError due to incorrect number of arguments from
    other TypeError
    """
    try:
        return fit_method(X, y, **kwargs)
    except TypeError:
        # fit takes only one argument
        return fit_method(X, **kwargs)


class TransformerPipeline(Pipeline):
    """
    Pipeline that expects all steps to be transformers taking a single X
    argument, an optional y argument, and having fit and transform methods.

    Code is copied from sklearn's Pipeline
    """

    def __init__(self, steps):
        names, estimators = zip(*steps)
        if len(dict(steps)) != len(steps):
            raise ValueError(
                "Provided step names are not unique: %s" % (names,))

        # shallow copy of steps
        self.steps = tosequence(steps)
        estimator = estimators[-1]

        for e in estimators:
            if (not (hasattr(e, "fit") or hasattr(e, "fit_transform")) or not
            hasattr(e, "transform")):
                raise TypeError("All steps of the chain should "
                                "be transforms and implement fit and transform"
                                " '%s' (type %s) doesn't)" % (e, type(e)))

        if not hasattr(estimator, "fit"):
            raise TypeError("Last step of chain should implement fit "
                            "'%s' (type %s) doesn't)"
                            % (estimator, type(estimator)))

    def _pre_transform(self, X, y=None, **fit_params):
        fit_params_steps = dict((step, {}) for step, _ in self.steps)
        for pname, pval in six.iteritems(fit_params):
            step, param = pname.split('__', 1)
            fit_params_steps[step][param] = pval
        Xt = X
        for name, transform in self.steps[:-1]:
            if hasattr(transform, "fit_transform"):
                Xt = _call_fit(transform.fit_transform,
                               Xt, y, **fit_params_steps[name])
            else:
                Xt = _call_fit(transform.fit,
                               Xt, y, **fit_params_steps[name]).transform(Xt)
        return Xt, fit_params_steps[self.steps[-1][0]]

    def fit(self, X, y=None, **fit_params):
        Xt, fit_params = self._pre_transform(X, y, **fit_params)
        _call_fit(self.steps[-1][-1].fit, Xt, y, **fit_params)
        return self

    def fit_transform(self, X, y=None, **fit_params):
        Xt, fit_params = self._pre_transform(X, y, **fit_params)
        if hasattr(self.steps[-1][-1], 'fit_transform'):
            return _call_fit(self.steps[-1][-1].fit_transform,
                             Xt, y, **fit_params)
        else:
            return _call_fit(self.steps[-1][-1].fit,
                             Xt, y, **fit_params).transform(Xt)


def make_transformer_pipeline(*steps):
    """Construct a TransformerPipeline from the given estimators.
    """
    return TransformerPipeline(_name_estimators(steps))


PY3 = sys.version_info[0] == 3
if PY3:
    string_types = text_type = str
else:
    string_types = basestring  # noqa
    text_type = unicode  # noqa


def _handle_feature(fea):
    """
    Convert 1-dimensional arrays to 2-dimensional column vectors.
    """
    if len(fea.shape) == 1:
        if isinstance(fea, dd.DataFrame):
            fea = da.stack([fea], axis=-1)
        else:
            fea = np.array([fea]).T

    return fea


def _build_transformer(transformers):
    if isinstance(transformers, list):
        transformers = make_transformer_pipeline(*transformers)
    return transformers


def _build_feature(columns, transformers, options={}):
    return (columns, _build_transformer(transformers), options)


def _get_feature_names(estimator, columns=None):
    """
    Attempt to extract feature names based on a given estimator
    """
    if hasattr(estimator, 'classes_'):
        return estimator.classes_
    elif hasattr(estimator, 'get_feature_names'):
        return estimator.get_feature_names(columns)
    return None


@contextlib.contextmanager
def add_column_names_to_exception(column_names):
    # Stolen from https://stackoverflow.com/a/17677938/356729
    try:
        yield
    except Exception as ex:
        if ex.args:
            msg = u'{}: {}'.format(column_names, ex.args[0])
        else:
            msg = text_type(column_names)
        ex.args = (msg,) + ex.args[1:]
        raise


class DataFrameMapper(BaseEstimator, TransformerMixin):
    """
    Map Pandas data frame column subsets to their own
    sklearn transformation.
    """

    def __init__(self, features, default=False, sparse=False, df_out=False,
                 input_df=False, df_out_dtype_transforms=None):
        """
        Params:

        features    a list of tuples with features definitions.
                    The first element is the pandas column selector. This can
                    be a string (for one column) or a list of strings.
                    The second element is an object that supports
                    sklearn's transform interface, or a list of such objects.
                    The third element is optional and, if present, must be
                    a dictionary with the options to apply to the
                    transformation. Example: {'alias': 'day_of_week'}

        default     default transformer to apply to the columns not
                    explicitly selected in the mapper. If False (default),
                    discard them. If None, pass them through untouched. Any
                    other transformer will be applied to all the unselected
                    columns as a whole, taken as a 2d-array.

        sparse      will return sparse matrix if set True and any of the
                    extracted features is sparse. Defaults to False.

        df_out      return a pandas data frame, with each column named using
                    the pandas column that created it (if there's only one
                    input and output) or the input columns joined with '_'
                    if there's multiple inputs, and the name concatenated with
                    '_1', '_2' etc if there's multiple outputs. NB: does not
                    work if *default* or *sparse* are true

        input_df    If ``True`` pass the selected columns to the transformers
                    as a pandas DataFrame or Series. Otherwise pass them as a
                    numpy array. Defaults to ``False``.
        """
        if df_out and (sparse or default):
            raise ValueError("Can not use df_out with sparse or default")

        self.features = features
        self.default = default
        self.sparse = sparse
        self.df_out = df_out
        self.input_df = input_df
        self.df_out_dtype_transforms = df_out_dtype_transforms

        # fitted
        self.fitted_features_ = None

    @staticmethod
    def _build(features, default):
        if isinstance(features, list):
            built_features = [_build_feature(*f) for f in features]
        else:
            built_features = features
        built_default = _build_transformer(default)

        return built_features, built_default

    def _get_col_subset(self, X, cols, input_df=False):
        """
        Get a subset of columns from the given table X.

        X       a Pandas dataframe; the table to select columns from
        cols    a string or list of strings representing the columns
                to select

        Returns a numpy array with the data from the selected columns
        """
        if isinstance(cols, string_types):
            return_vector = True
            cols = [cols]
        else:
            return_vector = False

        # Needed when using the cross-validation compatibility
        # layer for sklearn<0.16.0.
        # Will be dropped on sklearn-pandas 2.0.
        if isinstance(X, list):
            X = [x[cols] for x in X]
            X = pd.DataFrame(X)

        # elif isinstance(X, DataWrapper):
        #     X = X.df  # fetch underlying data

        if return_vector:
            t = X[cols[0]]
        else:
            t = X[cols]

        # return either a DataFrame/Series or a numpy array
        if input_df:
            return t
        else:
            return t.values

    def fit(self, X, y=None):
        built_features, built_default = self._build(self.features, self.default)

        fitted_features = []
        selected_columns = []

        for columns_def, transformers, options in built_features:
            logger.debug(f'columns:({columns_def}), transformers:({transformers}), options:({options})')
            if callable(columns_def):
                columns = columns_def(X)
            elif isinstance(columns_def, string_types):
                columns = [columns_def]
            else:
                columns = columns_def

            if isinstance(columns, (list, tuple)):
                columns = [c for c in columns if c not in selected_columns]

            fitted_features.append((columns, transformers, options))
            if columns is None or len(columns) <= 0:
                continue

            selected_columns += columns
            if transformers is not None:
                input_df = options.get('input_df', self.input_df)
                with add_column_names_to_exception(columns):
                    Xt = self._get_col_subset(X, columns, input_df)
                    _call_fit(transformers.fit, Xt, y)
                    # print(f'{transformers}:{Xt.dtypes}')

        # handle features not explicitly selected
        if built_default is not False and len(X.columns) > len(selected_columns):
            unselected_columns = [c for c in X.columns.to_list() if c not in selected_columns]
            if built_default is not None:
                with add_column_names_to_exception(unselected_columns):
                    Xt = self._get_col_subset(X, unselected_columns, self.input_df)
                    _call_fit(built_default.fit, Xt, y)
            fitted_features.append((unselected_columns, built_default, {}))

        self.fitted_features_ = fitted_features

        return self

    def transform(self, X):
        selected_columns = []
        transformed_columns = []
        extracted = []

        for columns, transformers, options in self.fitted_features_:
            if columns is None or len(columns) < 1:
                continue
            selected_columns += columns

            input_df = options.get('input_df', self.input_df)
            alias = options.get('alias')

            Xt = self._get_col_subset(X, columns, input_df)
            if transformers is not None:
                with add_column_names_to_exception(columns):
                    # print(f'before ---- {transformers}:{Xt.dtypes}')
                    Xt = transformers.transform(Xt)
                    # print(f'after ---- {transformers}:{pd.DataFrame(Xt).dtypes}')

            extracted.append(_handle_feature(Xt))

            transformed_columns += self.get_names(columns, transformers, Xt, alias)

        return self._to_transform_result(X, extracted, transformed_columns)

    def fit_transform(self, X, y=None, *fit_args):
        fitted_features = []
        selected_columns = []
        transformed_columns = []
        extracted = []

        built_features, built_default = self._build(self.features, self.default)
        for columns_def, transformers, options in built_features:
            if callable(columns_def):
                columns = columns_def(X)
            elif isinstance(columns_def, string_types):
                columns = [columns_def]
            else:
                columns = columns_def
            if isinstance(columns, (list, tuple)) and len(set(selected_columns).intersection(set(columns))) > 0:
                columns = [c for c in columns if c not in selected_columns]

            if columns is None or len(columns) < 1:
                continue

            fitted_features.append((columns, transformers, options))
            selected_columns += columns
            if logger.is_debug_enabled():
                logger.debug(f'fit_transform {len(columns)} columns with:\n{transformers}')

            input_df = options.get('input_df', self.input_df)
            alias = options.get('alias')

            Xt = self._get_col_subset(X, columns, input_df)
            if transformers is not None:
                with add_column_names_to_exception(columns):
                    if hasattr(transformers, 'fit_transform'):
                        Xt = _call_fit(transformers.fit_transform, Xt, y)
                    else:
                        _call_fit(transformers.fit, Xt, y)
                        Xt = transformers.transform(Xt)

            extracted.append(_handle_feature(Xt))
            if logger.is_debug_enabled():
                logger.debug(f'columns:{len(columns)}')
            transformed_columns += self.get_names(columns, transformers, Xt, alias)
            if logger.is_debug_enabled():
                logger.debug(f'transformed_names_:{len(transformed_columns)}')

        # handle features not explicitly selected
        if built_default is not False and len(X.columns) > len(selected_columns):
            unselected_columns = [c for c in X.columns.to_list() if c not in selected_columns]
            Xt = self._get_col_subset(X, unselected_columns, self.input_df)
            if built_default is not None:
                with add_column_names_to_exception(unselected_columns):
                    if hasattr(built_default, 'fit_transform'):
                        Xt = _call_fit(built_default.fit_transform, Xt, y)
                    else:
                        _call_fit(built_default.fit, Xt, y)
                        Xt = built_default.transform(Xt)
                transformed_columns += self.get_names(unselected_columns, built_default, Xt)
            else:
                # if not applying a default transformer, keep column names unmodified
                transformed_columns += unselected_columns
            extracted.append(_handle_feature(Xt))

            fitted_features.append((unselected_columns, built_default, {}))

        self.fitted_features_ = fitted_features

        return self._to_transform_result(X, extracted, transformed_columns)

    def get_names(self, columns, transformer, x, alias=None):
        """
        Return verbose names for the transformed columns.

        columns       name (or list of names) of the original column(s)
        transformer   transformer - can be a TransformerPipeline
        x             transformed columns (numpy.ndarray)
        alias         base name to use for the selected columns
        """
        # logger.debug(
        #     f'get_names: {isinstance(columns, list)}, len(columns):{len(columns)} columns:{columns}, alias:{alias}')
        if alias is not None:
            name = alias
        elif isinstance(columns, list):
            name = '_'.join(map(str, columns))
        else:
            name = columns
        num_cols = x.shape[1] if len(x.shape) > 1 else 1
        if num_cols > 1:
            # If there are as many columns as classes in the transformer,
            # infer column names from classes names.

            # If we are dealing with multiple transformers for these columns
            # attempt to extract the names from each of them, starting from the
            # last one
            # logger.debug(f'transformer:{transformer}')
            if isinstance(transformer, (TransformerPipeline, Pipeline)):
                inverse_steps = transformer.steps[::-1]
                estimators = (estimator for name, estimator in inverse_steps)
                names_steps = (_get_feature_names(e, columns) for e in estimators)
                names = next((n for n in names_steps if n is not None), None)

                if names is None and len(columns) == num_cols:
                    names = list(columns)
            # Otherwise use the only estimator present
            else:
                names = _get_feature_names(transformer, columns)

            if logger.is_debug_enabled():
                # logger.debug(f'names:{names}')
                logger.debug(f'names:{len(names)}')
            if names is not None and len(names) == num_cols:
                return list(names)  # ['%s_%s' % (name, o) for o in names]
            # otherwise, return name concatenated with '_1', '_2', etc.
            else:
                return [name + '_' + str(o) for o in range(num_cols)]
        else:
            return [name]

    def get_dtypes(self, extracted):
        dtypes_features = [self.get_dtype(ex) for ex in extracted]
        return [dtype for dtype_feature in dtypes_features
                for dtype in dtype_feature]

    def get_dtype(self, ex):
        if isinstance(ex, (np.ndarray, da.Array)) or _sparse.issparse(ex):
            return [ex.dtype] * ex.shape[1]
        elif isinstance(ex, (pd.DataFrame, dd.DataFrame)):
            return list(ex.dtypes)
        else:
            raise TypeError(type(ex))

    def _to_transform_result(self, X, extracted, transformed_columns):
        # combine the feature outputs into one array. at this point we lose track of which features
        # were created from which input columns, so it's assumed that that doesn't matter to the model.

        # no data transformed, raise a error
        if extracted is None or len(extracted) == 0:
            raise ValueError("No data output, ??? ")

        # If any of the extracted features is sparse, combine sparsely.
        # Otherwise, combine as normal arrays.
        if isinstance(X, dd.DataFrame):
            extracted = [a.values if isinstance(a, dd.DataFrame) else a for a in extracted]
            stacked = dex.hstack_array(extracted)
        elif any(_sparse.issparse(fea) for fea in extracted):
            stacked = _sparse.hstack(extracted).tocsr()
            # return a sparse matrix only if the mapper was initialized
            # with sparse=True
            if not self.sparse:
                stacked = stacked.toarray()
        else:
            stacked = np.hstack(extracted)

        if self.df_out:
            # if no rows were dropped preserve the original index,
            # otherwise use a new integer one
            if isinstance(X, dd.DataFrame):
                df_out = dd.from_dask_array(stacked, columns=transformed_columns, index=None)
            else:
                no_rows_dropped = len(X) == len(stacked)
                index = X.index if no_rows_dropped else None
                df_out = pd.DataFrame(stacked, columns=transformed_columns, index=index)

            # output different data types, if appropriate
            dtypes = self.get_dtypes(extracted)

            # preserve types
            for col, dtype, stype in zip(transformed_columns, dtypes, df_out.dtypes.tolist()):
                if dtype != stype:
                    if logger.is_debug_enabled():
                        logger.debug(f'convert {col} as {dtype} from {stype}')
                    df_out[col] = df_out[col].astype(dtype)
            df_out = self._dtype_transform(df_out)
            return df_out
        else:
            return stacked

    def _dtype_transform(self, df_out):
        if self.df_out_dtype_transforms is not None:
            for columns, dtype in self.df_out_dtype_transforms:
                if callable(columns):
                    columns = columns(df_out)
                if isinstance(columns, list) and len(columns) <= 0:
                    continue
                df_out[columns] = df_out[columns].astype(dtype)
        return df_out



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import pytest
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score, get_scorer
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OrdinalEncoder, StandardScaler
from sklearn.tree import ExtraTreeClassifier, DecisionTreeClassifier

from hypernets.tabular.column_selector import column_object_category_bool, column_number_exclude_timedelta
from hypernets.tabular.dataframe_mapper import DataFrameMapper
from hypernets.tabular.datasets import dsutils
from hypernets.tabular.ensemble import StackingEnsemble, AveragingEnsemble, GreedyEnsemble


def general_preprocessor():
    cat_transformer = Pipeline(
        steps=[('imputer_cat', SimpleImputer(strategy='constant')), ('encoder', OrdinalEncoder())])
    num_transformer = Pipeline(steps=[('imputer_num', SimpleImputer(strategy='mean')), ('scaler', StandardScaler())])

    preprocessor = DataFrameMapper(features=[(column_object_category_bool, cat_transformer),
                                             (column_number_exclude_timedelta, num_transformer)],
                                   input_df=True,
                                   df_out=True)
    return preprocessor


class TestStacking():
    def test_scorer(self):
        scorer = get_scorer('neg_log_loss')
        assert scorer

    def test_multi_estimators_binary(self):
        preprocessor = general_preprocessor()
        df = dsutils.load_bank().head(2000)
        y = df.pop('y')
        X = preprocessor.fit_transform(df)
        X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.7, random_state=9527)

        estimators = [RandomForestClassifier(), DecisionTreeClassifier(), ExtraTreeClassifier(), LogisticRegression(),
                      GradientBoostingClassifier()]

        rf_auc = self.get_auc(estimators[0], X_train, X_test, y_train, y_test)
        dt_auc = self.get_auc(estimators[1], X_train, X_test, y_train, y_test)
        et_auc = self.get_auc(estimators[2], X_train, X_test, y_train, y_test)
        lr_auc = self.get_auc(estimators[3], X_train, X_test, y_train, y_test)
        gb_auc = self.get_auc(estimators[4], X_train, X_test, y_train, y_test)

        ests = [RandomForestClassifier(), DecisionTreeClassifier(), LogisticRegression(), ExtraTreeClassifier(),
                LogisticRegression(), GradientBoostingClassifier()]

        greedy = GreedyEnsemble('binary', ests, need_fit=True, n_folds=5, ensemble_size=10)
        greedy_auc = self.get_auc(greedy, X_train, X_test, y_train, y_test)

        # save and load
        greedy.save('test_ensemble_output_')
        greedy = GreedyEnsemble.load('test_ensemble_output_')

        pred = greedy.predict(X_test)

        greedy = GreedyEnsemble('binary', ests, need_fit=True, n_folds=5, ensemble_size=10, scoring='accuracy')
        greedy_auc2 = self.get_auc(greedy, X_train, X_test, y_train, y_test)

        greedy = GreedyEnsemble('binary', ests, need_fit=True, n_folds=5, ensemble_size=10, scoring='roc_auc_ovo')
        greedy_auc3 = self.get_auc(greedy, X_train, X_test, y_train, y_test)

        avg = AveragingEnsemble('binary', ests, need_fit=True, n_folds=5, )
        avg_auc = self.get_auc(avg, X_train, X_test, y_train, y_test)

        avg_hard = AveragingEnsemble('binary', ests, need_fit=True, n_folds=5, method='hard')
        avg_auc_hard = self.get_auc(avg_hard, X_train, X_test, y_train, y_test)

        stacking_hard = StackingEnsemble('binary', ests, need_fit=True, n_folds=5, method='hard')
        stacking_auc_hard = self.get_auc(stacking_hard, X_train, X_test, y_train, y_test)

        stacking_soft = StackingEnsemble('binary', ests, need_fit=True, n_folds=5, method='soft')
        stacking_auc_soft = self.get_auc(stacking_soft, X_train, X_test, y_train, y_test)

        assert stacking_auc_soft

    def test_multi_estimators_multiclass(self):
        preprocessor = general_preprocessor()
        df = dsutils.load_glass_uci()
        y = df.pop(10)
        X = preprocessor.fit_transform(df)
        X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.7, random_state=9527)

        estimators = [RandomForestClassifier(), DecisionTreeClassifier(), ExtraTreeClassifier(), LogisticRegression(),
                      GradientBoostingClassifier()]

        rf_auc = self.get_auc(estimators[0], X_train, X_test, y_train, y_test)
        dt_auc = self.get_auc(estimators[1], X_train, X_test, y_train, y_test)
        et_auc = self.get_auc(estimators[2], X_train, X_test, y_train, y_test)
        lr_auc = self.get_auc(estimators[3], X_train, X_test, y_train, y_test)
        gb_auc = self.get_auc(estimators[4], X_train, X_test, y_train, y_test)

        ests = [RandomForestClassifier(), DecisionTreeClassifier(), ExtraTreeClassifier(), LogisticRegression(),
                GradientBoostingClassifier()]

        greedy = GreedyEnsemble('multiclass', ests, need_fit=True, n_folds=5, ensemble_size=10)
        greedy_auc = self.get_auc(greedy, X_train, X_test, y_train, y_test)

        avg = AveragingEnsemble('multiclass', ests, need_fit=True, n_folds=5, )
        avg_auc = self.get_auc(avg, X_train, X_test, y_train, y_test)

        avg_hard = AveragingEnsemble('multiclass', ests, need_fit=True, n_folds=5, method='hard')
        with pytest.raises(ValueError) as err:
            avg_auc_hard = self.get_auc(avg_hard, X_train, X_test, y_train, y_test)
            assert err

        stacking_hard = StackingEnsemble('multiclass', ests, need_fit=True, n_folds=5, method='hard')
        stacking_auc_hard = self.get_auc(stacking_hard, X_train, X_test, y_train, y_test)

        stacking_soft = StackingEnsemble('multiclass', ests, need_fit=True, n_folds=5, method='soft')
        stacking_auc_soft = self.get_auc(stacking_soft, X_train, X_test, y_train, y_test)

        assert stacking_auc_soft

    def get_auc(self, clf, X_train, X_test, y_train, y_test):
        clf.fit(X_train, y_train)
        proba = clf.predict_proba(X_test)
        if proba.shape[1] == 2:
            pred = proba[:, 1]
        else:
            pred = proba
        auc = roc_auc_score(y_test, pred, multi_class='ovo')
        return auc



# -*- coding:utf-8 -*-
"""

"""

import numpy as np
import hashlib
import threading
import contextlib
import queue
import copy
import time
from collections import OrderedDict
from .mutables import Mutable, MutableScope
from ..utils.common import generate_id, combinations
from ..utils import logging
from .random_state import get_random_state

logger = logging.get_logger(__name__)


class HyperNode(Mutable):
    def __init__(self, space=None, name=None):
        self._space = space if space is not None else get_default_space()
        Mutable.__init__(self, self._space.scope, name)
        self._space.add_node(self)

    def attach_to_space(self, space=None, name=None):
        self._space = space if space is not None else get_default_space()
        self.attach_to_scope(self._space.scope, name)
        self._space.add_node(self)

    @property
    def space(self):
        return self._space


class HyperSpace(Mutable):

    def __init__(self, scope=None, name=None):
        if scope is None:
            scope = MutableScope()
        Mutable.__init__(self, scope, name)
        self.edges = set()
        self.modules = set()
        self.hyper_params = set()
        self._inputs = set()
        self._outputs = set()
        self._assigned_params_stack = []
        self._is_compiled = False
        self.space_id = generate_id()

    @property
    def type(self):
        return 'DAG'

    def as_default(self):
        return _default_space_stack.get_controller(self)

    @property
    def all_assigned(self):
        all_assigned = self.traverse(lambda m: m.all_assigned, direction='forward', discard_isolated_node=False)
        return all_assigned

    @property
    def assigned_params_stack(self):
        return self._assigned_params_stack

    def push_assigned_param(self, param):
        self._assigned_params_stack.append(param)

    @property
    def params_iterator(self):
        visited = {}
        while not self.all_assigned:
            for p in self.get_unassigned_params():
                if not p.assigned:
                    if visited.get(p):
                        visited[p] += 1
                        if visited[p] > 10:
                            return
                            # raise RuntimeError('Too many attempts to get assignable params')
                    else:
                        visited[p] = 1
                        yield p

    def add_node(self, node):
        if isinstance(node, ModuleSpace):
            self.modules.add(node)
        elif isinstance(node, ParameterSpace):
            self.hyper_params.add(node)
        else:
            raise ValueError(f"Not supported node:{node}")
        self.__dict__[node.id] = node

    def compile(self, deepcopy=True):
        if deepcopy:
            space = copy.deepcopy(self)
        else:
            space = self
        space._compile_space()
        return space

    def _compile_space(self):
        assert not self._is_compiled, 'HyperSpace does not allow to compile repeatedly.'
        space_out = []
        counter = 0
        start_ts = time.time()

        def compile_module(module):
            module.compile()
            nonlocal counter
            counter += 1
            if len(self.get_outputs(module)) <= 0:
                space_out.append(module)
            return True

        self.traverse(compile_module, direction='forward')
        end_ts = time.time()
        if logger.is_info_enabled():
            logger.debug(f'Compile Space: compiled {counter} modules in {end_ts - start_ts} seconds.')
        self._is_compiled = True
        self._outputs = set(space_out)

    def forward(self, inputs=None):
        counter = []
        start_ts = time.time()

        def forward_module(module):
            input_modules = self.get_inputs(module)
            if len(input_modules) <= 0:
                # The input module of space
                module.forward(inputs)
            elif len(input_modules) == 1:
                module.forward(inputs=input_modules[0].output)
            else:
                module.forward(inputs=[m.output for m in input_modules])
            counter.append(0)
            return True

        self.traverse(forward_module, direction='forward')
        end_ts = time.time()
        if logger.is_info_enabled():
            logger.debug(f'Forward Space: forwarded {len(counter)} modules in {end_ts - start_ts} seconds.')
        outputs = [output.output for output in self.get_outputs()]
        return outputs

    def compile_and_forward(self, inputs=None, deepcopy=True):
        space = self.compile(deepcopy)
        outputs = space.forward(inputs)
        return space, outputs

    def traverse(self, fn, direction='forward', start_modules=[], discard_isolated_node=True):
        if direction == 'forward':
            fn_inputs = self.get_inputs
            fn_outputs = self.get_outputs
        elif direction == 'backward':
            fn_inputs = self.get_outputs
            fn_outputs = self.get_inputs
        else:
            raise ValueError(f'Not supported direction:{direction}')

        standby = queue.Queue()
        visited = set()
        finished = set()
        if start_modules is None or len(start_modules) <= 0:
            start_modules = fn_inputs(discard_isolated_node=discard_isolated_node)
        for m in start_modules:
            standby.put(m)
            visited.add(m)

        while not standby.empty():
            m_todo = standby.get()
            inputs = fn_inputs(m_todo)
            ready = True
            for mi in inputs:
                if mi not in finished:
                    ready = False
                    break

            if not ready:
                visited.remove(m_todo)
                continue

            is_continues = fn(m_todo)

            if not is_continues:
                return False
            finished.add(m_todo)
            for m in fn_outputs(m_todo):
                if not m in visited:
                    standby.put(m)
                    visited.add(m)
        return True

    def connect(self, from_module, to_module):
        self.edges.add((from_module, to_module))

    def disconnect(self, from_module, to_module):
        found = False
        for f, t in self.edges:
            if f == from_module and t == to_module:
                found = True
                break
        if len(self.edges & {(from_module, to_module)}) == 1:
            self.edges.remove((from_module, to_module))

    def disconnect_all(self, module):
        found = set()
        for f, t in self.edges:
            if f == module or t == module:
                found.add((f, t))
        for f, t in found:
            self.edges.remove((f, t))

    def reroute_to(self, old_module, new_module):
        assert isinstance(new_module, (list, ModuleSpace))
        found = set()
        for f, t in self.edges:
            if t == old_module:
                found.add((f, t))

        for f, t in found:
            self.edges.remove((f, t))
            if isinstance(new_module, ModuleSpace):
                self.edges.add((f, new_module))
            else:
                for m in new_module:
                    self.edges.add((f, m))

    def reroute_from(self, old_module, new_module):
        found = set()
        assert isinstance(new_module, (list, ModuleSpace))

        for f, t in self.edges:
            if f == old_module:
                found.add((f, t))

        for f, t in found:
            self.edges.remove((f, t))
            if isinstance(new_module, ModuleSpace):
                self.edges.add((new_module, t))
            else:
                for m in new_module:
                    self.edges.add((m, t))

    def replace_route(self, old_module, new_module):
        self.reroute_to(old_module, new_module)
        self.reroute_from(old_module, new_module)

    def is_isolated_module(self, module):
        assert module is not None
        for from_module, to_module in self.edges:
            if module == from_module or module == to_module:
                return False
        return True

    def get_sub_graph_outputs(self, module_in_subgraph):
        return self.get_sub_graph_end_modules(module_in_subgraph, direction='forward')

    def get_sub_graph_inputs(self, module_in_subgraph):
        return self.get_sub_graph_end_modules(module_in_subgraph, direction='backward')

    def get_sub_graph_end_modules(self, module_in_subgraph, direction='forward'):
        assert isinstance(module_in_subgraph, (ModuleSpace, list))
        if isinstance(module_in_subgraph, list):
            module_in_subgraph = module_in_subgraph[0]

        if direction == 'forward':  # get outputs
            get_upstream = self.get_outputs
            get_downstream = self.get_inputs
        elif direction == 'backward':  # get inputs
            get_upstream = self.get_inputs
            get_downstream = self.get_outputs

        else:
            raise ValueError(f'Not supported direction:{direction}')

        standby = queue.Queue()
        visited = {module_in_subgraph}
        finished = {module_in_subgraph}
        end_modules = set()
        if len(get_upstream(module_in_subgraph)) <= 0:
            end_modules.add(module_in_subgraph)
        for m in get_upstream(module_in_subgraph):
            standby.put(m)
            visited.add(m)
        for m in get_downstream(module_in_subgraph):
            standby.put(m)
            visited.add(m)

        while not standby.empty():
            m_todo = standby.get()
            m_downstream = get_downstream(m_todo)
            ready = True
            for mi in m_downstream:
                if mi not in finished:
                    standby.put(mi)
                    visited.add(mi)
                    ready = False
                    break

            if not ready:
                if len(visited & {m_todo}) > 0:
                    visited.remove(m_todo)
                continue

            finished.add(m_todo)
            m_upstream = get_upstream(m_todo)
            if len(m_upstream) <= 0:
                end_modules.add(m_todo)
                continue

            for m in m_upstream:
                if not m in visited:
                    standby.put(m)
                    visited.add(m)
        return sorted(end_modules, key=lambda m: m.id)

    def set_inputs(self, modules):
        assert modules is not None
        assert isinstance(modules, (ModuleSpace, list))
        if isinstance(modules, ModuleSpace):
            self._inputs = {modules}
        else:
            assert len(modules) > 0
            assert all([isinstance(m, ModuleSpace) for m in modules])
            self._inputs = set(modules)

    def set_outputs(self, modules):
        assert modules is not None
        assert isinstance(modules, (ModuleSpace, list))
        if isinstance(modules, ModuleSpace):
            self._outputs = {modules}
        else:
            assert len(modules) > 0
            assert all([isinstance(m, ModuleSpace) for m in modules])
            self._outputs = set(modules)

    def get_inputs(self, module=None, discard_isolated_node=True):
        inputs = set()
        if module is None:
            if len(self._inputs) > 0:
                return sorted(self._inputs, key=lambda m: m.id)

            for m in self.modules:
                has_in = False
                has_out = False
                for from_module, to_module in self.edges:
                    if m == from_module:
                        has_out = True
                    if m == to_module:
                        has_in = True
                        break
                if not has_in:
                    if has_out or (not discard_isolated_node and m.path == ''):
                        inputs.add(m)

            if len(inputs) == 0:
                if len(self.modules) == 1:
                    inputs = self.modules.copy()
                else:
                    raise ValueError('Graph is not connected.')

        else:
            for from_module, to_module in self.edges:
                if module == to_module:
                    inputs.add(from_module)

        return sorted(inputs, key=lambda m: m.id)

    def get_outputs(self, module=None, discard_isolated_node=True):
        outputs = set()
        if module is None:
            if len(self._outputs) > 0:
                return sorted(self._outputs, key=lambda m: m.id)

            for m in self.modules:
                has_in = False
                has_out = False
                for from_module, to_module in self.edges:
                    if m == from_module:
                        has_out = True
                        break
                    if m == to_module:
                        has_in = True
                if not has_out:
                    if has_in or (not discard_isolated_node and m.path == ''):
                        outputs.add(m)
            if len(outputs) == 0:
                if len(self.modules) == 1:
                    outputs = self.modules.copy()
                else:
                    raise ValueError('Graph is not connected.')
        else:
            for from_module, to_module in self.edges:
                if module == from_module:
                    outputs.add(to_module)
        return sorted(outputs, key=lambda m: m.id)

    def random_sample(self):
        for hp in self.params_iterator:
            hp.random_sample()

    def get_unassigned_params(self, traverse_direction='forward'):
        assignables = []

        def append_params(m):
            ps = m.get_assignable_params()
            for p in ps:
                if p not in assignables:
                    assignables.append(p)
            return True

        self.traverse(append_params, direction=traverse_direction, discard_isolated_node=False)
        return assignables

    def get_assigned_params(self):
        assert self.all_assigned
        return self._assigned_params_stack

    def get_assigned_param_values(self, traverse_direction='forward'):
        ps = self.get_assigned_params()
        return {p.id: p.value for p in ps}

    def get_all_params(self):
        all = list(self.hyper_params)
        return all

    def params_summary(self, only_assignable=True, line_width=60, LR='\n'):
        outputs = []
        outputs.append(f'\n{(line_width + 2) * "-"}')
        if only_assignable:
            params = self.get_assigned_params()
        else:
            params = self.get_all_params()

        for i, hp in enumerate(params):
            outputs.append(
                f'({i}) {hp.alias}:{(line_width - len(str(i) + "() " + hp.alias + str(hp.value))) * " "}{hp.value}')
        outputs.append(f'{(line_width + 2) * "-"}')
        return LR.join(outputs)

    @property
    def signature(self):
        assert self.all_assigned
        labels = [p.label for p in self._assigned_params_stack]
        key = ';'.join(labels)
        md5 = hashlib.md5(key.encode('utf-8')).hexdigest()
        return md5

    @property
    def vectors(self):
        assert self.all_assigned
        vectors = [p.value2numeric(p.value) for p in self._assigned_params_stack]
        return vectors

    def assign_by_vectors(self, vectors):
        i = 0
        for p in self.params_iterator:
            if not p.is_mutable:
                p.random_sample()
                continue
            if i >= len(vectors):
                raise ValueError('`vector` and `space` does not match.')
            p.assign(p.numeric2value(vectors[i]))
            i += 1
        if len(vectors) != i:
            raise ValueError('`vector` and `space` does not match.')

    @property
    def combinations(self):
        count = 1
        for hp in self.hyper_params:
            count *= hp.choice_num
        return count

    def _repr_html_(self):
        html = '''<table border="1" class="dataframe">
<thead>
<tr style="text-align: right;">
  <th>key</th>
  <th>value</th>
</tr>
</thead>
<tbody>'''
        html += f'''<tr>
  <td>signature</td>
  <td>{self.signature}</td>
</tr>
<tr>
  <td>vectors</td>
  <td>{self.vectors}</td>
</tr>'''
        params = self.get_assigned_params()
        for i, hp in enumerate(params):
            html += f'''<tr>
  <td>{i}-{hp.alias}</td>
  <td>{hp.value}</td>
</tr>
<tr>'''

        html += '''  </tbody>
</table>
</div>'''
        return html


class DefaultStack(threading.local):
    def __init__(self):
        super(DefaultStack, self).__init__()
        self.enforce_nesting = True
        self.stack = []

    def get_default(self):
        return self.stack[-1] if len(self.stack) >= 1 else None

    def reset(self):
        self.stack = []

    def is_cleared(self):
        return not self.stack

    @contextlib.contextmanager
    def get_controller(self, default):
        self.stack.append(default)
        try:
            yield default
        finally:
            if self.stack:
                if self.enforce_nesting:
                    if self.stack[-1] is not default:
                        raise AssertionError(
                            "Nesting violated for default stack of %s objects" %
                            type(default))
                    self.stack.pop()
                else:
                    self.stack.remove(default)


class DefaultSpaceStack(DefaultStack):
    def __init__(self):
        super(DefaultSpaceStack, self).__init__()
        self._global_default_space = None

    def get_default(self):
        """Override that returns a global default if the stack is empty."""
        default = super(DefaultSpaceStack, self).get_default()
        if default is None:
            default = self._global_default()
        return default

    def _global_default(self):
        if self._global_default_space is None:
            self._global_default_space = HyperSpace()
        return self._global_default_space

    def reset(self):
        super(DefaultSpaceStack, self).reset()
        self._global_default_space = None


_default_space_stack = DefaultSpaceStack()


def get_default_space():
    return _default_space_stack.get_default()


class ParameterSpace(HyperNode):
    def __init__(self, space=None, name=None, random_state=None):
        HyperNode.__init__(self, space, name)
        self._assigned = False
        self._value = None
        self.random_state = random_state if random_state is not None else get_random_state()
        self.references = set()

    @property
    def is_mutable(self):
        return True

    @property
    def type(self):
        return 'Param'

    @property
    def assigned(self):
        return self._assigned

    @property
    def value(self):
        return self._value

    @property
    def config_keys(self):
        raise NotImplementedError

    @property
    def label(self):
        vs = [str(self.__dict__[key]) for key in self.config_keys]
        return f"{self.id}-{'-'.join(vs)}"

    def value2numeric(self, value):
        raise NotImplementedError

    def numeric2value(self, numeric):
        raise NotImplementedError

    def random_sample(self, assign=True):
        value = self._random_sample()
        if assign:
            self.assign(value)
        return value

    def _random_sample(self):
        raise NotImplementedError

    def assign(self, value):
        assert not self._assigned
        self._check(value)
        self._assigned = True
        self._value = value
        if self.is_mutable:
            self.space.push_assigned_param(self)
        for m in self.references:
            m.update()

    def attach(self, mutable, alias=None):
        self.references.add(mutable)
        if alias is not None:
            all = [] if self.alias is None else [self.alias]
            all.append(mutable.name + '.' + alias)
            self.alias = ','.join(all)

    def detach(self, mutable):
        self.references.remove(mutable)

    def _check(self, value):
        pass

    def same_config(self, other):
        if self.__class__ == other.__class__ and self.alias == other.alias:
            for key in self.config_keys:
                if self.__dict__[key] != other.__dict__[key]:
                    return False
            return True
        else:
            return False

    def expansion(self, sample_num):
        raise NotImplementedError

    @property
    def choice_num(self):
        if self.is_mutable:
            return self._get_choice_num()
        else:
            return 1

    def _get_choice_num(self):
        raise NotImplementedError


class Int(ParameterSpace):
    def __init__(self, low, high, step=1, random_state=None, space=None, name=None):
        ParameterSpace.__init__(self, space, name, random_state)
        assert isinstance(low, int) and isinstance(high, int), '`low` and `high` must be a int.'
        assert low < high, '`low` must less than `high`.'
        self.low = low
        self.high = high
        self.step = step

    def _random_sample(self):
        value = self.random_state.randint(self.low, self.high)
        if self.step is not None:
            all = np.arange(self.low, self.high + self.step, step=self.step)
            value = all[np.abs(all - value).argmin()]
        return value

    def _check(self, value):
        assert value >= self.low and value <= self.high

    def value2numeric(self, value):
        return value

    def numeric2value(self, numeric):
        return numeric

    @property
    def config_keys(self):
        return ['low', 'high', 'step']

    def expansion(self, sample_num):
        p = self._get_choice_num()
        if sample_num > p or sample_num <= 0:
            sample_num = p
        samples = []
        values = []
        while len(samples) < sample_num:
            v = self._random_sample()
            if v in values:
                continue
            sample = copy.deepcopy(self)
            sample.assign(v)
            samples.append(sample)
            values.append(v)

        return sorted(samples, key=lambda s: s.value)

    def _get_choice_num(self):
        p = self.high - self.low
        if self.step is not None:
            p = len(np.arange(self.low, self.high + self.step, step=self.step)) - 1
        return p


class Real(ParameterSpace):
    def __init__(self, low, high, q=None, prior="uniform", step=0.01, max_expansion=100,
                 random_state=None, space=None, name=None):
        ParameterSpace.__init__(self, space, name, random_state)
        low = float(low)
        high = float(high)
        assert low < high, '`low` must less than `high`.'
        self.low = low
        self.high = high
        self.q = q
        self.prior = prior
        self.step = float(step)
        self.max_expansion = max_expansion

    def _random_sample(self):
        if self.prior == "uniform":
            assert self.high >= self.low, 'Upper bound must be larger than lower bound'
            value = self.random_state.uniform(self.low, self.high)
            self._check(value)
        elif self.prior == "log_uniform":
            assert self.low > 0, 'Lower bound must be positive'
            value = np.exp(self.random_state.uniform(self.low, self.high))
        elif self.prior == "q_uniform":
            assert self.q is not None, 'q cannot be None'
            value = np.clip(
                np.round(self.random_state.uniform(self.low, self.high) / self.q) * self.q, self.low, self.high)
            self._check(value)
        else:
            raise ValueError(f'Not supported prior:{self.prior}')

        if self.step is not None:
            if self.prior == 'log_uniform':
                all = np.round(np.arange(np.exp(self.low), np.exp(self.high) + self.step, step=self.step), 8)
                value = all[np.abs(all - value).argmin()]
                if value > np.exp(self.high):
                    value = np.exp(self.high)
            else:
                all = np.round(np.arange(self.low, self.high + self.step, step=self.step), 8)
                value = all[np.abs(all - value).argmin()]
                if value > self.high:
                    value = self.high
        return value

    def _check(self, value):
        if self.prior == "log_uniform":
            assert value >= np.exp(self.low) and value <= np.exp(self.high)
        else:
            assert value >= self.low and value <= self.high

    @property
    def config_keys(self):
        return ['low', 'high', 'q', 'prior', 'step']

    def value2numeric(self, value):
        return value

    def numeric2value(self, numeric):
        return numeric

    def expansion(self, sample_num):
        if sample_num <= 0:
            sample_num = self.max_expansion
        sample_num = min(sample_num, self._get_choice_num())
        values = []
        samples = []
        while len(samples) < sample_num:
            v = self._random_sample()
            if v in values:
                continue
            sample = copy.deepcopy(self)
            sample.assign(v)
            samples.append(sample)
            values.append(v)
        return sorted(samples, key=lambda s: s.value)

    def _get_choice_num(self):
        p = self.max_expansion
        if self.step is not None:
            if self.prior == 'log_uniform':
                p = len(np.arange(np.exp(self.low), np.exp(self.high) + self.step, step=self.step))
            else:
                p = len(np.arange(self.low, self.high + self.step, step=self.step))
        return p


class Choice(ParameterSpace):
    def __init__(self, options, random_state=None, space=None, name=None):
        ParameterSpace.__init__(self, space, name, random_state)
        assert isinstance(options, list), '`options` must be a List.'
        assert len(options) > 0, '`options` contains at least one item.'
        self.options = options

    @property
    def is_mutable(self):
        return len(self.options) > 1

    def _random_sample(self):
        index = self.random_state.choice(range(len(self.options)))
        return self.options[index]

    def _check(self, value):
        assert value in self.options

    @property
    def config_keys(self):
        return ['options']

    def value2numeric(self, value):
        return self.options.index(value)

    def numeric2value(self, numeric):
        return self.options[numeric]

    def expansion(self, sample_num=0):
        samples = []
        for option in self.options:
            sample = copy.deepcopy(self)
            sample.assign(option)
            samples.append(sample)
        return samples

    def _get_choice_num(self):
        return len(self.options)


class MultipleChoice(ParameterSpace):
    def __init__(self, options, num_chosen_most=0, num_chosen_least=1, random_state=None,
                 space=None, name=None):
        ParameterSpace.__init__(self, space, name, random_state)
        assert isinstance(options, list), '`options` must be a List.'
        assert len(options) >= num_chosen_least, f'`options` contains at least {num_chosen_least} item.'
        self.options = options
        self.num_chosen_most = num_chosen_most
        self.num_chosen_least = num_chosen_least

    def _random_sample(self):
        high = self.num_chosen_most
        if high <= 0:
            high = len(self.options)
        indices = self.random_state.choice(range(0, len(self.options)),
                                           self.random_state.randint(self.num_chosen_least, high + 1), False)
        values = [self.options[index] for index in sorted(indices)]
        return values

    def _check(self, value):
        assert isinstance(value, list)
        assert len(value) >= self.num_chosen_least, f'value contains at least {self.num_chosen_least} item.'
        assert (self.num_chosen_most == 0 or self.num_chosen_most <= len(self.options))
        assert all([v in self.options for v in value])

    @property
    def config_keys(self):
        return ['options', 'num_chosen_most', 'num_chosen_least']

    def value2numeric(self, value):
        numeric = int(''.join(['1' if v in value else '0' for v in self.options]), 2)
        return numeric

    def numeric2value(self, numeric):
        bin = np.binary_repr(numeric, len(self.options))

        values = []
        for i in range(len(bin)):
            if bin[i] == '1':
                values.append(self.options[i])
        return values

    def expansion(self, sample_num):
        c = self._get_choice_num()
        if sample_num > c or sample_num <= 0:
            sample_num = c
        values = []
        samples = []
        while len(values) < sample_num:
            v = self._random_sample()
            if v in values:
                continue
            sample = copy.deepcopy(self)
            sample.assign(v)
            samples.append(sample)
            values.append(v)
        return samples

    def _get_choice_num(self):
        return int(combinations(len(self.options), self.num_chosen_most, 1))


class Bool(Choice):
    def __init__(self, random_state=None, space=None, name=None):
        Choice.__init__(self, [False, True], random_state, space, name)


class Constant(ParameterSpace):
    def __init__(self, value, space=None, name=None):
        ParameterSpace.__init__(self, space, name)
        self.assign(value)

    @property
    def is_mutable(self):
        return False

    @property
    def config_keys(self):
        return ['_value']


class Dynamic(ParameterSpace):
    def __init__(self, lambda_fn, space=None, name=None, **param_dict):
        ParameterSpace.__init__(self, space, name)
        self._lambda_fn = lambda_fn
        self._param_dict = {}

        for n, p in param_dict.items():
            if isinstance(p, Dynamic):
                raise ValueError('Dynamic cannot be nested.')
            self._param_dict[n] = p
            p.attach(self, n)

        self.update()

    @property
    def is_mutable(self):
        return False

    def update(self):
        if all(p.assigned for p in self._param_dict.values()):
            args = {name: p.value for name, p in self._param_dict.items()}
            value = self._lambda_fn(**args)
            self.assign(value)

    @property
    def param_dict(self):
        return self._param_dict


class Cascade(ParameterSpace):
    def __init__(self, lambda_fn, space=None, name=None, **param_dict):
        ParameterSpace.__init__(self, space, name)
        self._lambda_fn = lambda_fn
        self._param_dict = {}

        for n, p in param_dict.items():
            self._param_dict[n] = p
            p.attach(self, n)

        self.update()

    @property
    def is_mutable(self):
        return False

    @property
    def assigned(self):
        if self.value is not None:
            if isinstance(self.value, ParameterSpace):
                return self.value.assigned
            else:
                return True

    def update(self):
        if all(p.assigned for p in self._param_dict.values()):
            args = {name: p.value for name, p in self._param_dict.items()}
            name, value = self._lambda_fn(args, self.space)
            assert isinstance(value, ParameterSpace), 'The value of `Cascade` must be a ParameterSpace.'
            for m in self.references:
                if isinstance(m, ModuleSpace):
                    m.add_parameters(**{name: value})
                elif isinstance(m, ParameterSpace):
                    value.attach(m)
            self.assign(value)

    @property
    def param_dict(self):
        return self._param_dict


class ModuleSpace(HyperNode):
    def __init__(self, space=None, name=None, **hyperparams):
        HyperNode.__init__(self, space, name)
        self._hyper_params = OrderedDict()
        self._is_params_ready = False
        self._is_compiled = False
        self._output = None
        self.space.add_node(self)
        self.add_parameters(**hyperparams)
        self.update()

    def __call__(self, *args, **kwargs):
        assert len(args) > 0
        m = args[0]
        assert m is not None, 'The module to be connected cannot be None.'
        assert isinstance(m, (ModuleSpace, list))
        if isinstance(m, ModuleSpace):
            self.space.connect(m, self)
        else:
            for mi in m:
                assert mi is not None, 'The module to be connected cannot be None.'
                self.space.connect(mi, self)
        return self

    def connect(self, module_or_list):
        if isinstance(module_or_list, ModuleSpace):
            self.space.connect(self, module_or_list)
        elif isinstance(module_or_list, list):
            assert len(module_or_list) > 0, f'module_or_list contains at least 1 Module.'
            assert all([isinstance(m, ModuleSpace) for m in module_or_list]), 'module_or_list can only contain Module.'
            for m in module_or_list:
                self.space.connect(self, m)
        else:
            raise ValueError(f'module_or_list is neither Module nor List.')
        return self

    @property
    def type(self):
        return 'Module'

    @property
    def param_values(self):
        return {name: p.value for name, p in self._hyper_params.items()}

    @property
    def hyper_params(self):
        return self._hyper_params

    def get_assignable_params(self):
        assignables = []
        for name, p in self._hyper_params.items():
            if isinstance(p, Constant):
                continue
            elif isinstance(p, (Dynamic, Cascade)):
                for dp in p.param_dict.values():
                    if not isinstance(dp, (Dynamic, Cascade)) and not dp in assignables:
                        assignables.append(dp)
            else:
                if not p in assignables:
                    assignables.append(p)
        return assignables

    def get_all_params(self):
        all = []
        for name, p in self._hyper_params.items():
            if not p in all:
                all.append(p)
            if isinstance(p, (Dynamic, Cascade)):
                for dp in p.param_dict.values():
                    if not dp in all:
                        all.append(dp)
        return all

    @property
    def output(self):
        return self._output

    @property
    def is_compiled(self):
        return self._is_compiled

    @property
    def is_params_ready(self):
        return self._is_params_ready

    @property
    def all_assigned(self):
        for hp in self._hyper_params.values():
            if not hp.assigned:
                return False
        return True

    def compile(self):
        if not self.is_compiled:
            self._compile()
            self._is_compiled = True

    def _compile(self):
        raise NotImplementedError

    def forward(self, inputs=None):
        self._output = self._forward(inputs)
        return self._output

    def _forward(self, inputs):
        raise NotImplementedError

    def compile_and_forward(self, inputs=None):
        self.compile()
        return self.forward(inputs)

    def _on_params_ready(self):
        pass

    def add_parameters(self, **hyperparameters):
        for name, param in hyperparameters.items():
            if not isinstance(param, ParameterSpace):
                param = Constant(param)
            if self._hyper_params.get(name) is not None:
                raise ValueError(f'Parameter `{name}` has existed.')
            self._hyper_params[name] = param
            param.attach(self, name)

    def update(self):
        if all(p.assigned for p in self._hyper_params.values()):
            self._is_params_ready = True
            self._on_params_ready()



from sklearn.model_selection import train_test_split

from hypernets.examples.plain_model import PlainModel, PlainSearchSpace
from hypernets.experiment import make_experiment
from hypernets.experiment.compete import StepNames
from hypernets.tabular.datasets import dsutils


def test_experiment_with_blood_simple():
    df = dsutils.load_blood()
    experiment = make_experiment(PlainModel, df, target='Class', search_space=PlainSearchSpace())
    estimator = experiment.run(max_trials=3)
    print(estimator)
    assert estimator is not None


def test_experiment_with_blood_down_sample():
    df = dsutils.load_blood()
    experiment = make_experiment(PlainModel, df, target='Class', search_space=PlainSearchSpace(),
                                 down_sample_search=True, down_sample_search_size=0.1,
                                 down_sample_search_time_limit=300, down_sample_search_max_trials=10,
                                 # log_level='info',
                                 )
    estimator = experiment.run(max_trials=3)
    print(estimator)
    assert estimator is not None


def test_experiment_with_blood_full_features():
    df = dsutils.load_blood()
    target = 'Class'
    df_train, df_test = train_test_split(df, train_size=0.8, random_state=335)
    df_test.pop(target)

    experiment = make_experiment(PlainModel, df, target=target, search_space=PlainSearchSpace(),
                                 test_data=df_test,
                                 feature_generation=True,
                                 collinearity_detection=True,
                                 drift_detection=True,
                                 feature_selection=True,
                                 down_sample_search=True, down_sample_search_size=0.2,
                                 feature_reselection=True,
                                 pseudo_labeling=True,
                                 random_state=335,
                                 early_stopping_time_limit=1200,
                                 # log_level='info',
                                 )
    estimator = experiment.run(max_trials=3)
    print(estimator)
    assert estimator is not None

    step_names = [step[0] for step in estimator.steps]
    assert step_names == [StepNames.DATA_CLEAN, StepNames.MULITICOLLINEARITY_DETECTION, 'estimator']



import queue
import time

import grpc

from hypernets.utils import logging
from .proto import spec_pb2_grpc
from .proto.spec_pb2 import SearchRequest, SearchResponse, PingMessage

logger = logging.get_logger(__name__)


class SearchDriverClient(object):
    _instances = {}
    callbacks = []

    class TrialItemWrapper(object):
        def __init__(self, code, search_id, trial_no,
                     space_id, space_file, model_file):
            super(SearchDriverClient.TrialItemWrapper, self).__init__()

            self.code = code
            self.search_id = search_id
            try:
                self.trial_no = int(trial_no)
            except ValueError:
                self.trial_no = None
            self.space_id = space_id
            self.space_file = space_file
            self.model_file = model_file

            self.success = False
            self.reward = None
            self.message = None

        def is_ok(self):
            return self.code == SearchResponse.OK

        def is_waiting(self):
            return self.code == SearchResponse.WAITING

        def is_finished(self):
            return self.code == SearchResponse.FINISHED

        def to_request(self):
            msg = SearchRequest(search_id=self.search_id,
                                space_id=self.space_id,
                                trial_no=str(self.trial_no) if self.trial_no is not None else '',
                                success=self.success,
                                reward=self.reward if self.reward is not None else 0.0,
                                message=self.message if self.message else '')
            return msg

        @classmethod
        def from_response(cls, msg):
            item = cls(code=msg.code,
                       search_id=msg.search_id,
                       trial_no=msg.trial_no,
                       space_id=msg.space_id,
                       space_file=msg.space_file,
                       model_file=msg.model_file)
            return item

    def __init__(self, server, search_id):
        super(SearchDriverClient, self).__init__()
        self.channel = grpc.insecure_channel(server)
        self.stub = spec_pb2_grpc.SearchDriverStub(self.channel)

        self.server = server
        self.search_id = search_id

    def close(self):
        self.channel.close()

    def ping(self, wait=False, message=None):

        def default_message():
            try:
                import socket
                hostname = socket.gethostname()
                return f'from-{hostname}'
            except Exception:
                return 'ping'

        def do_ping():
            req = PingMessage(message=message if message else default_message())
            res = self.stub.ping(req)
            return res.message

        if wait:
            start_at = time.time()
            result = ''
            while True:
                try:
                    result = do_ping()
                    done_at = time.time()
                    print(' connected with %.3f seconds.' % (done_at - start_at))
                    break
                except grpc.RpcError:
                    print('.', end='', flush=True)
                    time.sleep(1)

            return result
        else:
            return do_ping()

    def search(self, search_id):

        ack_queue = queue.Queue()
        running = True

        def fire_request():
            # fire first msg,
            msg = SearchRequest(search_id=search_id,
                                trial_no='',
                                space_id='',
                                success=False,
                                reward=0.0,
                                message='')
            yield msg

            # fire response msg and get next trial
            while running:
                try:
                    r = ack_queue.get(block=False)
                    if r is None:
                        break
                    msg = r.to_request()
                    yield msg
                except queue.Empty:
                    time.sleep(0.1)
                except Exception:
                    import traceback
                    traceback.print_exc()
                    break

        try:
            response = self.stub.search(fire_request())
            for res in response:
                item = SearchDriverClient.TrialItemWrapper.from_response(res)
                try:
                    result = yield item
                    ack_queue.put(result)
                except Exception as e:
                    msg = f'{e.__class__.__name__}: {e}'
                    logger.error(msg)
        except grpc.RpcError as e:
            import traceback
            trace_detail = traceback.format_exc()
            try:
                msg = f'RpcError {self.server} {e.__class__.__name__}: {e.code()}'
                logger.error(msg)
            except Exception:
                msg = f'RpcError {self.server} {e.__class__.__name__}:\n'
                logger.error(msg + trace_detail)
            yield None
        running = False



import inspect
import re
from collections import OrderedDict

_KEY_PARAMETERS = 'Parameters'


class DocLens(object):
    def __init__(self, doc):
        self.doc_ = doc  # original doc

        self.synopsis, self.sections = self._parse_doc(doc)

    @property
    def parameters(self):
        if _KEY_PARAMETERS in self.sections.keys():
            return self._parse_parameters(self.sections[_KEY_PARAMETERS])
        else:
            return OrderedDict()

    @parameters.setter
    def parameters(self, params):
        assert isinstance(params, OrderedDict)

        content = '\n'.join([f'{k}{v}' for k, v in params.items()])
        self.sections[_KEY_PARAMETERS] = content

    def render(self):
        result = []
        if self.synopsis is not None and len(self.synopsis) > 0:
            result.append(self.synopsis)

        for name, doc in self.sections.items():
            result.append(f'{name}\n------\n{doc}')

        return '\n'.join(result)

    def merge_parameters(self, other, exclude=None):
        if exclude is None:
            exclude = []

        params = self.parameters
        for k, v in other.parameters.items():
            if k not in exclude and k not in params.keys():
                params[k] = v

        return params

    @staticmethod
    def _parse_doc(doc):
        """
        split pydoc into synopsis and sections dict
        """
        doc = inspect.cleandoc(doc)

        sa = re.split(r'\n[-=]+\n', doc)
        if len(sa) <= 1:
            return doc, {}

        synopsis = ''
        sections = OrderedDict()
        section_name = None

        for s in sa[:-1]:
            i = s.rfind('\n')
            if i >= 0:
                content = s[:i]
                next_section_name = s[i + 1:]
            else:
                content = ''
                next_section_name = s

            if section_name is None:
                synopsis = content
            else:
                sections[section_name] = s[:i]
            section_name = next_section_name
        sections[section_name] = sa[-1]

        return synopsis, sections

    @staticmethod
    def _parse_parameters(doc):
        """
        split pydoc into parameter dict
        """
        params = OrderedDict()
        detail = None
        for line in doc.split('\n'):
            m = re.search(r'^(\w+)(.*)', line)
            if m:
                name, synopsis = m.groups()
                detail = [synopsis]
                params[name] = detail
            elif detail is not None:
                detail.append(line)

        result = OrderedDict()
        for name, detail in params.items():
            result[name] = '\n'.join(detail)

        return result



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import copy
import sys
import time

import dask
import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from lightgbm.sklearn import LGBMClassifier
from sklearn import model_selection as sksel
from sklearn.metrics import roc_auc_score, matthews_corrcoef, make_scorer

from hypernets.tabular import dask_ex as dex
from hypernets.tabular.column_selector import column_object_category_bool, column_number_exclude_timedelta
from hypernets.utils import logging
from .cfg import TabularCfg as cfg
from .general import general_preprocessor, general_estimator

logger = logging.getLogger(__name__)

is_os_windows = sys.platform.find('win') >= 0
roc_auc_scorer = make_scorer(roc_auc_score, greater_is_better=True, needs_threshold=True)
matthews_corrcoef_scorer = make_scorer(matthews_corrcoef)


class FeatureSelectionCallback():
    def on_round_start(self, round_no, features, ):
        pass

    def on_round_end(self, round_no, auc, features, remove_features, elapsed):
        pass

    def on_remove_shift_variable(self, shift_score, remove_features):
        pass

    def on_task_break(self, round_no, auc, features):
        pass

    def on_task_finished(self, round_no, auc, features):
        pass


def feature_selection(X_train, X_test,
                      remove_shift_variable=True,
                      variable_shift_threshold=0.7,
                      variable_shift_scorer=None,
                      auc_threshold=0.55,
                      min_features=10, remove_size=0.1,
                      preprocessor=None,
                      estimator=None, sample_balance=True, max_test_samples=None, cv=5, random_state=9527,
                      copy_data=False,
                      callbacks=None):
    logger.info('Feature selection to try to eliminate the concept drift.')
    if copy_data:
        X_train = X_train.copy() if isinstance(X_train, dd.DataFrame) else copy.deepcopy(X_train)
        X_test = X_test.copy() if isinstance(X_test, dd.DataFrame) else copy.deepcopy(X_test)

    scores = None
    if remove_shift_variable:
        scores = covariate_shift_score(X_train, X_test, scorer=variable_shift_scorer)
        remain_features = []
        remove_features = []
        for col, score in scores.items():
            if score <= variable_shift_threshold:
                remain_features.append(col)
            else:
                remove_features.append(col)
                logger.info(f'Remove shift variables:{col},  score:{score}')
        if len(remain_features) < X_train.shape[1]:
            X_train = X_train[remain_features]
            X_test = X_test[remain_features]
        if callbacks is not None:
            for callback in callbacks:
                callback.on_remove_shift_variable(scores, remove_features)
    round = 1
    history = []
    while True:
        start_time = time.time()
        if callbacks is not None:
            for callback in callbacks:
                callback.on_round_start(round_no=round, features=X_train.columns.to_list())
        logger.info(f'\nRound: {round}\n')
        detector = DriftDetector(preprocessor, estimator, random_state)
        detector.fit(X_train, X_test, sample_balance=sample_balance, max_test_samples=max_test_samples, cv=cv)
        logger.info(f'AUC:{detector.auc_}, Features:{detector.feature_names_}')
        elapsed = time.time() - start_time
        history.append({'auc': detector.auc_,
                        'n_features': len(detector.feature_names_),
                        'removed_features': [],
                        'feature_names': detector.feature_names_,
                        'feature_importances': detector.feature_importances_,
                        'elapsed': elapsed
                        })

        if detector.auc_ <= auc_threshold:
            logger.info(
                f'AUC:{detector.auc_} has dropped below the threshold:{auc_threshold}, feature selection is over.')
            if callbacks is not None:
                for callback in callbacks:
                    callback.on_task_finished(round_no=round, auc=detector.auc_, features=detector.feature_names_)
            return detector.feature_names_, history, scores

        indices = np.argsort(detector.feature_importances_)
        if indices.shape[0] <= min_features:
            logger.info(f'The number of remaining features is insufficient to continue remove features. '
                        f'AUC:{detector.auc_} '
                        f'Remaining features:{detector.feature_names_}')
            if callbacks is not None:
                for callback in callbacks:
                    callback.on_task_break(round_no=round, auc=detector.auc_, features=detector.feature_names_)
            return detector.feature_names_, history, scores

        removes = int(indices.shape[0] * remove_size)
        if removes <= 0:
            logger.info(f'The number of remaining features is insufficient to continue remove features. '
                        f'AUC:{detector.auc_} '
                        f'Remaining features:({len(detector.feature_names_)}) / {detector.feature_names_}')
            if callbacks is not None:
                for callback in callbacks:
                    callback.on_task_break(round_no=round, auc=detector.auc_, features=detector.feature_names_)
            return detector.feature_names_, history, scores

        if (indices.shape[0] - removes) < min_features:
            removes = indices.shape[0] - min_features

        remain_features = list(np.array(detector.feature_names_)[indices[:-removes]])
        remove_features = list(set(detector.feature_names_) - set(remain_features))
        history[-1]['removed_features'] = remove_features

        logger.info(f'Removed features: {remove_features}')
        X_train = X_train[remain_features]
        X_test = X_test[remain_features]
        if callbacks is not None:
            for callback in callbacks:
                callback.on_round_end(round_no=round, auc=detector.auc_, features=detector.feature_names_,
                                      remove_features=remove_features, elapsed=elapsed)
        round += 1


class DriftDetector():
    def __init__(self, preprocessor=None, estimator=None, random_state=9527):
        self.preprocessor = preprocessor
        self.estimator_ = estimator
        self.random_state = random_state
        self.auc_ = None
        self.feature_names_ = None
        self.feature_importances_ = None
        self.fitted = False

    def fit(self, X_train, X_test, sample_balance=True, max_test_samples=None, cv=5):
        logger.info('Fit data for concept drift detection')
        assert X_train.shape[1] == X_test.shape[1], 'The number of columns in X_train and X_test must be the same.'
        assert len(set(X_train.columns.to_list()) - set(
            X_test.columns.to_list())) == 0, 'The name of columns in X_train and X_test must be the same.'

        if dex.exist_dask_dataframe(X_train, X_test):
            import dask_ml.model_selection as dsel
            train_shape, test_shape = dask.compute(X_train.shape, X_test.shape)
            iterators = dsel.KFold(n_splits=cv, shuffle=True, random_state=1001)
        else:
            train_shape, test_shape = X_train.shape, X_test.shape
            iterators = sksel.StratifiedKFold(n_splits=cv, shuffle=True, random_state=1001)

        if max_test_samples is not None and max_test_samples < test_shape[0]:
            X_test, _ = dex.train_test_split(X_test, train_size=max_test_samples, random_state=self.random_state)
            test_shape = (max_test_samples, test_shape[0])
        if sample_balance:
            if test_shape[0] > train_shape[0]:
                X_test, _ = dex.train_test_split(X_test, train_size=train_shape[0], random_state=self.random_state)
                test_shape = (train_shape[0], test_shape[1])
            elif test_shape[0] < train_shape[0]:
                X_train, _ = dex.train_test_split(X_train, train_size=test_shape[0], random_state=self.random_state)
                train_shape = (test_shape[0], train_shape[1])

        target_col = '__drift_detection_target__'

        if hasattr(X_train, 'insert'):
            X_train.insert(0, target_col, 0)
        else:
            X_train[target_col] = 0

        if hasattr(X_test, 'insert'):
            X_test.insert(0, target_col, 1)
        else:
            X_test[target_col] = 1

        X_merge = dex.concat_df([X_train, X_test], repartition=True)
        y = X_merge.pop(target_col)

        logger.info('Preprocessing...')

        if self.preprocessor is None:
            self.preprocessor = general_preprocessor(X_merge)
        X_merge = self.preprocessor.fit_transform(X_merge)

        self.feature_names_ = X_merge.columns.to_list()
        self.feature_importances_ = []
        auc_all = []
        importances = []
        estimators = []

        if dex.is_dask_dataframe(X_merge):
            X_values, y_values = X_merge.to_dask_array(lengths=True), y.to_dask_array(lengths=True)
        else:
            # X_values, y_values = X_merge.values, y.values
            X_values, y_values = X_merge, y

        for n_fold, (train_idx, valid_idx) in enumerate(iterators.split(X_values, y_values)):
            logger.info(f'Fold:{n_fold + 1}')
            if dex.is_dask_dataframe(X_merge):
                x_train_fold, y_train_fold = X_values[train_idx], y_values[train_idx]
                x_val_fold, y_val_fold = X_values[valid_idx], y_values[valid_idx]
                x_train_fold = dex.array_to_df(x_train_fold, meta=X_merge)
                x_val_fold = dex.array_to_df(x_val_fold, meta=X_merge)
            else:
                x_train_fold, y_train_fold = X_values.iloc[train_idx], y_values.iloc[train_idx]
                x_val_fold, y_val_fold = X_values.iloc[valid_idx], y_values.iloc[valid_idx]

            estimator = general_estimator(X_merge, self.estimator_)
            kwargs = {}
            # if isinstance(estimator, dask_xgboost.XGBClassifier):
            #     kwargs['eval_set'] = [(x_val_fold.compute(), y_val_fold.compute())]
            #     kwargs['early_stopping_rounds'] = 10
            if isinstance(estimator, LGBMClassifier):
                kwargs['eval_set'] = (x_val_fold, y_val_fold)
                kwargs['early_stopping_rounds'] = 10
                kwargs['verbose'] = 0
            estimator.fit(x_train_fold, y_train_fold, **kwargs)
            proba = estimator.predict_proba(x_val_fold)[:, 1]
            if dex.is_dask_dataframe(X_merge):
                y_val_fold, proba = dex.compute(y_val_fold, proba, traverse=False)
            auc = roc_auc_score(y_val_fold, proba)
            logger.info(f'auc: {auc}')

            auc_all.append(auc)
            estimators.append(estimator)
            importances.append(estimator.feature_importances_)

        self.estimator_ = estimators
        self.auc_ = np.mean(auc_all)
        self.feature_importances_ = np.mean(importances, axis=0)
        self.fitted = True
        X_test.pop(target_col)
        X_train.pop(target_col)
        return self

    def predict_proba(self, X):
        assert self.fitted, 'Please fit it first.'

        X = X.copy() if dex.is_dask_dataframe(X) else copy.deepcopy(X)
        cat_cols = column_object_category_bool(X)
        num_cols = column_number_exclude_timedelta(X)
        # X.loc[:, cat_cols + num_cols] = self.preprocessor.transform(X)
        Xt = self.preprocessor.transform(X)
        diff_cols = set(X.columns.tolist()) - set(cat_cols + num_cols)
        if diff_cols:
            # X.loc[:, cat_cols + num_cols] = Xt
            X = dex.concat_df([X[diff_cols], Xt[cat_cols + num_cols]], axis=1)
        else:
            X = Xt

        oof_proba = []
        for i, estimator in enumerate(self.estimator_):
            proba = estimator.predict_proba(X)[:, 1]
            oof_proba.append(proba)
        if dex.is_dask_dataframe(X):
            proba = da.mean(dex.hstack_array(oof_proba), axis=1)
        else:
            proba = np.mean(oof_proba, axis=0)
        return proba

    def train_test_split(self, X, y, test_size=0.25, remain_for_train=0.3):
        if dex.exist_dask_object(X, y):
            return self.train_test_split_by_dask(X, y, test_size=test_size, remain_for_train=remain_for_train)

        assert remain_for_train < 1.0 and remain_for_train >= 0, '`remain_for_train` must be < 1.0 and >= 0.'
        if isinstance(test_size, float):
            assert test_size < 1.0 and test_size > 0, '`test_size` must be < 1.0 and > 0.'
            test_size = int(X.shape[0] * test_size)
        assert isinstance(test_size, int), '`test_size` can only be int or float'
        split_size = int(test_size + test_size * remain_for_train)
        assert split_size < X.shape[0], \
            'test_size+test_size*remain_for_train must be less than the number of samples in X.'

        proba = self.predict_proba(X)
        sorted_indices = np.argsort(proba)
        target = '__train_test_split_y__'
        X.insert(0, target, y)

        if remain_for_train == 0:
            X_train = X.iloc[sorted_indices[:-test_size]]
            X_test = X.iloc[sorted_indices[-test_size:]]
            y_train = X_train.pop(target)
            y_test = X_test.pop(target)
            return X_train, X_test, y_train, y_test
        else:
            X_train_1 = X.iloc[sorted_indices[:-split_size]]
            X_mixed = X.iloc[sorted_indices[-split_size:]]
            X_train_2, X_test = dex.train_test_split(X_mixed, test_size=test_size, shuffle=True,
                                                     random_state=self.random_state)
            X_train = pd.concat([X_train_1, X_train_2], axis=0)
            y_train = X_train.pop(target)
            y_test = X_test.pop(target)
            return X_train, X_test, y_train, y_test

    def train_test_split_by_dask(self, X, y, test_size=0.25, remain_for_train=0.3):
        x_shape = dex.compute(X.shape)[0]
        assert remain_for_train < 1.0 and remain_for_train >= 0, '`remain_for_train` must be < 1.0 and >= 0.'
        if isinstance(test_size, float):
            assert test_size < 1.0 and test_size > 0, '`test_size` must be < 1.0 and > 0.'
            test_size = int(x_shape[0] * test_size)
        assert isinstance(test_size, int), '`test_size` can only be int or float'
        split_size = int(test_size + test_size * remain_for_train)
        assert split_size < x_shape[0], \
            'test_size+test_size*remain_for_train must be less than the number of samples in X.'

        X = X.copy()
        proba = self.predict_proba(X)
        sorted_indices = np.argsort(proba.compute())
        target = '__train_test_split_y__'
        X[target] = y

        X_values = X.to_dask_array(lengths=True)
        if remain_for_train == 0:
            X_train = dex.array_to_df(X_values[sorted_indices[:-test_size]], meta=X)
            X_test = dex.array_to_df(X_values[sorted_indices[-test_size:]], meta=X)
            y_train = X_train.pop(target)
            y_test = X_test.pop(target)
            return X_train, X_test, y_train, y_test
        else:
            X_train_1 = dex.array_to_df(X_values[sorted_indices[:-split_size]], meta=X)
            X_mixed = dex.array_to_df(X_values[sorted_indices[-split_size:]], meta=X)
            X_train_2, X_test = dex.train_test_split(X_mixed, test_size=test_size, shuffle=True,
                                                     random_state=self.random_state)
            X_train = dex.concat_df([X_train_1, X_train_2], axis=0)
            y_train = X_train.pop(target)
            y_test = X_test.pop(target)
            return X_train, X_test, y_train, y_test


def covariate_shift_score(X_train, X_test, scorer=None, cv=None, copy_data=True):
    assert all(isinstance(x, (pd.DataFrame, dd.DataFrame)) for x in (X_train, X_test)), \
        'X_train and X_test must be a pandas or dask DataFrame.'

    assert len(set(X_train.columns.to_list()) - set(
        X_test.columns.to_list())) == 0, 'The columns in X_train and X_test must be the same.'
    if scorer is None:
        scorer = roc_auc_scorer
    if copy_data:
        train = X_train.copy() if dex.is_dask_dataframe(X_train) else copy.deepcopy(X_train)
        test = X_test.copy() if dex.is_dask_dataframe(X_test) else copy.deepcopy(X_test)
    else:
        train = X_train
        test = X_test

    # Set target value
    target_col = '__hypernets_csd__target__'
    if hasattr(train, 'insert'):
        train.insert(0, target_col, 0)
    else:
        train[target_col] = 0
    if hasattr(test, 'insert'):
        test.insert(0, target_col, 1)
    else:
        test[target_col] = 1

    X_merge = dex.concat_df([train, test], axis=0)
    y = X_merge.pop(target_col)
    if dex.is_dask_dataframe(X_merge):
        y = y.compute()

    logger.info('Preprocessing...')
    # Preprocess data: imputing and scaling
    preprocessor = general_preprocessor(X_merge)
    X_merge = preprocessor.fit_transform(X_merge)
    if dex.is_dask_dataframe(X_merge):
        X_merge = X_merge.persist()

    # Calculate the shift score for each column separately.
    scores = {}
    logger.info('Scoring...')
    if dex.is_dask_dataframe(X_merge) or cfg.joblib_njobs in {0, 1}:
        for c in X_merge.columns:
            x = X_merge[[c]]
            if dex.is_dask_dataframe(X_merge):
                x = x.compute()
            score = _shift_score(x, y, scorer, cv)
            logger.info(f'column:{c}, score:{score}')
            scores[c] = score
    else:
        col_parts = [X_merge[[c]] for c in X_merge.columns]
        options = dict(backend='multiprocessing') if is_os_windows else dict(prefer='processes')
        pss = Parallel(n_jobs=cfg.joblib_njobs, **options)(delayed(_shift_score)(x, y, scorer, cv) for x in col_parts)
        scores = {k: v for k, v in zip(X_merge.columns.to_list(), pss)}
        logger.info(f'scores: {scores}')

    return scores


def _shift_score(x, y, scorer, cv):
    model = general_estimator(x)
    if cv:
        score_ = sksel.cross_val_score(model, X=x, y=y, verbose=0, scoring=scorer, cv=cv)
        score = np.mean(score_)
    else:
        mixed_x_train, mixed_x_test, mixed_y_train, mixed_y_test = \
            sksel.train_test_split(x, y, test_size=0.3, random_state=9527, stratify=y)
        model.fit(mixed_x_train, mixed_y_train,
                  eval_set=(mixed_x_test, mixed_y_test),
                  early_stopping_rounds=20,
                  verbose=False)
        score = scorer(model, mixed_x_test, mixed_y_test)

    return score



import json

import numpy as np
import pandas as pd
from dask import dataframe as dd

from hypernets.tabular import column_selector as col_se


def get_data_character(hyper_model, X_train, y_train, X_eval=None, y_eval=None, X_test=None, task=None):

	dtype2usagetype = {'object':'str', 'int64':'int', 'float64':'float', 'datetime64[ns]':'date', 'timedelta64[ns]':'date'}
	
	task, _ = hyper_model.infer_task_type(y_train) #This line is just used to test

	if isinstance(y_train, pd.core.series.Series):
		datatype_y = dtype2usagetype[str(y_train.dtypes)]

		Missing_y = y_train.isnull().sum().tolist()
		Unique_y = len(y_train.unique())

		if task == 'binary':
			Freq_y = y_train.value_counts()[0].tolist()
		else:
			Freq_y = None
		
		if task == 'regression':
			max_y = max(y_train)
			min_y = min(y_train)
			mean_y = pd.Series.mean(y_train)
			Stdev_y = y_train.std()
		else:
			max_y = None
			min_y = None
			mean_y = None
			Stdev_y = None

		if task == 'regression':
			Freq_y = None
			disc_y_num = None
			cont_y_num = 10
			interval = (max_y-min_y)/cont_y_num
			intervals = np.linspace(min_y, max_y, cont_y_num+1)
			Freqs_y = pd.cut(y_train, intervals).value_counts(sort=False)  # [0:cont_y_num]
			count = list(Freqs_y)
			region = list(map(lambda x: [x.left, x.right], list(Freqs_y.keys())))
			target_distribution = {'count': count, 'region': region}
		else:
			cont_y_num = None
			disc_y_num = 10
			target_distribution = dict(y_train.value_counts()[0:disc_y_num])
			for key in target_distribution:
				target_distribution[key] = int(target_distribution[key])
		
		shape_x_train = list(X_train.shape)
		shape_y_train = list(y_train.shape)
		if X_eval is None:
			shape_x_eval = []
		else:
			shape_x_eval = list(X_eval.shape)

		if y_eval is None:
			shape_y_eval = []
		else:
			shape_y_eval = list(y_eval.shape)

		if X_test is None:
			shape_x_test = []
		else:
			shape_x_test = list(X_test.shape)

	else:
		datatype_y = dtype2usagetype[str(y_train.dtype)]

		Missing_y = y_train.isnull().compute().tolist().count(True)
		Unique_y = len(y_train.unique().compute().tolist())

		if task == 'binary':	
			Freq_y = y_train.value_counts().compute().tolist()[0]
		else:
			Freq_y = None

		if task == 'regression':
			max_y = y_train.max().compute().tolist()
			min_y = y_train.min().compute().tolist()
			mean_y = y_train.mean().compute().tolist()
			Stdev_y = y_train.std().compute().tolist()
		else:
			max_y = None
			min_y = None
			mean_y = None
			Stdev_y = None
		
		if task == 'regression':
			Freq_y = None
			disc_y_num = None
			cont_y_num = 10
			interval = (max_y-min_y)/cont_y_num
			intervals = np.linspace(min_y, max_y, cont_y_num+1)
			Freqs_y = pd.cut(y_train, intervals).value_counts()[0:cont_y_num]
			count = list(Freqs_y)
			region = list(map(lambda x: [x.left.tolist(), x.right.tolist()], list(Freqs_y.keys())))
			target_distribution = {'count':count, 'region':region}
		else:
			cont_y_num = None
			disc_y_num = 10
			target_distribution = dict(dd.compute(y_train.value_counts())[0][0:disc_y_num])
			for key in target_distribution:
				target_distribution[key] = int(target_distribution[key])

		shape_x_train = list(X_train.shape)
		for idx, num in enumerate(shape_x_train):
			if isinstance(num, int):
				continue
			else:
				shape_x_train[idx] = num.compute()

		shape_y_train = list(map(lambda x: x.compute(), list(y_train.shape)))

		if X_eval is None:
			shape_x_eval = []
		else:
			shape_x_eval = list(X_eval.shape)
			for idx, num in enumerate(shape_x_eval):
				if isinstance(num, int):
					continue
				else:
					shape_x_eval[idx] = num.compute()
		
		if y_eval is None:
			shape_y_eval = []
		else:
			shape_y_eval = list(y_eval.shape)
			for idx, num in enumerate(shape_y_eval):
				if isinstance(num, int):
					continue
				else:
					shape_y_eval[idx] = num.compute()

		if X_test is None:
			shape_x_test = []
		else:
			shape_x_test = list(X_test.shape)
			for idx, num in enumerate(shape_x_test):
				if isinstance(num, int):
					continue
				else:
					shape_x_test[idx] = num.compute()
		
	data_character = {
		'experimentType': 'base',
		'target':{
			'name':'y',
			'taskType':task,
			'freq':Freq_y,
			'unique':Unique_y,
			'missing':Missing_y,
			'mean':mean_y, 
			'min':min_y,
			'max':max_y,
			'stdev':Stdev_y, 
			'dataType':datatype_y
		},
		'targetDistribution': target_distribution,
		'datasetShape':{
			'X_train':shape_x_train, 
			'y_train':shape_y_train, 
			'X_eval':shape_x_eval, 
			'y_eval':shape_y_eval, 
			'X_test':shape_x_test
		}
		}

	return data_character

def get_x_data_character(X_train, get_step):

	cnt_x_all = len(col_se.column_all(X_train))
	cnt_x_date = len(col_se.column_all_datetime(X_train))
	cnt_x_category = len(col_se.column_object_category_bool(X_train))
	cnt_x_num = len(col_se.column_number(X_train))

	try:
		kwargs = get_step('feature_generation').transformer_kwargs
		if kwargs['text_cols'] != None:
			cnt_x_text = len(kwargs['text_cols'])
			cnt_x_all -= len(col_se.column_text(X_train)) - len(kwargs['text_cols'])
		else:
			cnt_x_text = len(col_se.column_text(X_train))
			
		if kwargs['latlong_cols'] != None:
			cnt_x_latlong = len(kwargs['latlong_cols'])
			cnt_x_all += len(kwargs['latlong_cols'])
		else:
			cnt_x_latlong = 0
	except:
		cnt_x_text = len(col_se.column_text(X_train))
		cnt_x_latlong = 0

	cnt_x_others = cnt_x_all - cnt_x_date - cnt_x_category - cnt_x_num - cnt_x_text - cnt_x_latlong

	x_types = {
		'experimentType': 'compete',
		'featureDistribution':{
		'nContinuous':cnt_x_num, 
		'nText':cnt_x_text,
		'nDatetime':cnt_x_date,
		'nCategorical':cnt_x_category,
		'nLocation':cnt_x_latlong,
		'nOthers':cnt_x_others
		}
	}

	return x_types




from hypernets.core.callbacks import SummaryCallback
from hypernets.examples.plain_model import PlainModel, PlainSearchSpace
from hypernets.examples.plain_model import train_heart_disease
from hypernets.searchers import make_searcher
from hypernets.tabular.sklearn_ex import MultiLabelEncoder


class DaskPlainModel(PlainModel):
    def _get_estimator(self, space_sample):
        from hypernets.tabular import dask_ex as dex

        estimator = super()._get_estimator(space_sample)

        return dex.wrap_local_estimator(estimator)


def create_plain_model(reward_metric='auc', optimize_direction='max',
                       with_encoder=False, with_dask=False):
    search_space = PlainSearchSpace(enable_dt=True, enable_lr=True, enable_nn=False)
    searcher = make_searcher('random', search_space_fn=search_space, optimize_direction=optimize_direction)

    encoder = MultiLabelEncoder if with_encoder else None
    cls = DaskPlainModel if with_dask else PlainModel
    hyper_model = cls(searcher=searcher, reward_metric=reward_metric, callbacks=[SummaryCallback()],
                      transformer=encoder)

    return hyper_model


def test_train_heart_disease():
    train_heart_disease(cv=False, max_trials=5)


def test_train_heart_disease_with_cv():
    train_heart_disease(cv=True, max_trials=5)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from sklearn.model_selection import train_test_split

from hypernets.utils import logging
from . import Experiment

logger = logging.get_logger(__name__)


class GeneralExperiment(Experiment):
    def __init__(self, hyper_model, X_train, y_train, X_eval=None, y_eval=None, X_test=None, eval_size=0.3,
                 task=None, id=None, callbacks=None, random_state=9527):
        super(GeneralExperiment, self).__init__(hyper_model, X_train, y_train, X_eval=X_eval,
                                                y_eval=y_eval, X_test=X_test, eval_size=eval_size, task=task,
                                                id=id, callbacks=callbacks, random_state=random_state)

    def train(self, hyper_model, X_train, y_train, X_test, X_eval=None, y_eval=None, **kwargs):
        """Run an experiment

        Arguments
        ---------
        max_trials :

        """
        self.step_start('data split')
        if X_eval is None or y_eval is None:
            stratify = y_train
            if self.task == 'regression':
                stratify = None
            X_train, X_eval, y_train, y_eval = train_test_split(X_train, y_train, test_size=self.eval_size,
                                                                random_state=self.random_state, stratify=stratify)
        self.step_end(output={'X_train.shape': X_train.shape,
                              'y_train.shape': y_train.shape,
                              'X_eval.shape': X_eval.shape,
                              'y_eval.shape': y_eval.shape,
                              'X_test.shape': None if X_test is None else X_test.shape})

        self.step_start('search')
        hyper_model.search(X_train, y_train, X_eval, y_eval, **kwargs)
        best_trial = hyper_model.get_best_trial()
        self.step_end(output={'best_trial': best_trial})

        self.step_start('load estimator')
        self.estimator = hyper_model.load_estimator(best_trial.model_file)
        self.step_end(output={'estimator': self.estimator})

        return self.estimator



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from lightgbm import LGBMRegressor
import numpy as np
from ..utils import logging

logger = logging.get_logger(__name__)


class MetaLearner(object):
    def __init__(self, history, dataset_id, trial_store):
        self.trial_store = trial_store
        self.dataset_id = dataset_id
        self.history = history
        self.regressors = {}
        self.store_history = {}

        if logger.is_info_enabled():
            logger.info(f'Initialize Meta Learner: dataset_id:{dataset_id}')

    def new_sample(self, space_sample):
        self.fit(space_sample.signature)

    def fit(self, space_signature):

        features = self.extract_features_and_labels(space_signature)
        x = []
        y = []
        for features, label in features:
            x.append(features)
            y.append(label)

        store_history = self.store_history.get(space_signature)

        if self.trial_store is not None and store_history is None:
            trials = self.trial_store.get_all(self.dataset_id, space_signature)
            store_x = []
            store_y = []
            for t in trials:
                store_x.append(t.space_sample_vectors)
                store_y.append(t.reward)
            store_history = (store_x, store_y)
            self.store_history[space_signature] = store_history

        if store_history is None:
            store_history = ([], [])

        store_x, store_y = store_history
        x = x + store_x
        y = y + store_y
        if len(x) >= 2:
            regressor = LGBMRegressor()
            regressor.fit(x, y)
            #  if logger.is_info_enabled():
            #      logger.info(regressor.predict(x))
            self.regressors[space_signature] = regressor

    def predict(self, space_sample, default_value=np.inf):
        regressor = self.regressors.get(space_sample.signature)
        if regressor is not None:
            score = regressor.predict([space_sample.vectors])
        else:
            score = default_value
        return score

    def extract_features_and_labels(self, signature):
        features = [(t.space_sample.vectors, t.reward) for t in self.history.trials if
                    t.space_sample.signature == signature]
        return features
